from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.metagenome.constants import MetagenomeDeliveryType
from cg.services.orders.validation.order_types.metagenome.models.sample import MetagenomeSample


class MetagenomeOrder(OrderWithSamples[MetagenomeSample]):
    delivery_type: MetagenomeDeliveryType



import logging
from datetime import datetime

from cg.constants import DataDelivery, GenePanelMasterList, Priority, Workflow
from cg.constants.constants import CustomerId
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.models.orders.sample_base import SexEnum, StatusEnum
from cg.services.orders.constants import ORDER_TYPE_WORKFLOW_MAP
from cg.services.orders.lims_service.service import OrderLimsService
from cg.services.orders.storing.constants import MAF_ORDER_ID
from cg.services.orders.storing.service import StoreOrderService
from cg.services.orders.validation.order_types.fastq.models.order import FastqOrder
from cg.services.orders.validation.order_types.fastq.models.sample import FastqSample
from cg.store.models import ApplicationVersion, Case, CaseSample, Customer, Order, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class StoreFastqOrderService(StoreOrderService):
    """Storing service for FASTQ orders."""

    def __init__(self, status_db: Store, lims_service: OrderLimsService):
        self.status_db = status_db
        self.lims = lims_service

    def store_order(self, order: FastqOrder) -> dict:
        """Submit a batch of samples for FASTQ delivery."""
        project_data, lims_map = self.lims.process_lims(
            samples=order.samples,
            ticket=order._generated_ticket_id,
            order_name=order.name,
            workflow=Workflow.RAW_DATA,
            customer=order.customer,
            delivery_type=DataDelivery(order.delivery_type),
            skip_reception_control=order.skip_reception_control,
        )
        self._fill_in_sample_ids(samples=order.samples, lims_map=lims_map)
        new_samples: list[Sample] = self.store_order_data_in_status_db(order=order)
        return {"records": new_samples, "project": project_data}

    def store_order_data_in_status_db(self, order: FastqOrder) -> list[Sample]:
        """
        Store all order data in the Status database for a FASTQ order. Return the samples.
        The stored data objects are:
        - Order
        - Samples
        - One Case per sample
        - For each Sample, a relationship between the sample and its Case
        - For each non-tumour WGS Sample, a MAF Case and a relationship between the Sample and the
        MAF Case
        """
        db_order: Order = self._create_db_order(order=order)
        new_samples = []
        with self.status_db.session.no_autoflush:
            for sample in order.samples:
                db_case: Case = self._create_db_case_for_sample(
                    sample=sample, customer=db_order.customer, order=order
                )
                db_sample: Sample = self._create_db_sample(
                    sample=sample,
                    order_name=order.name,
                    ticket_id=str(db_order.ticket_id),
                    customer=db_order.customer,
                )
                self._create_maf_case(db_sample=db_sample, db_order=db_order, db_case=db_case)
                case_sample: CaseSample = self.status_db.relate_sample(
                    case=db_case,
                    sample=db_sample,
                    status=StatusEnum.unknown,
                    should_deliver_sample=True,
                )
                self.status_db.add_multiple_items_to_store([db_sample, case_sample, db_case])
                new_samples.append(db_sample)
                db_order.cases.append(db_case)
        self.status_db.add_item_to_store(db_order)
        self.status_db.commit_to_store()
        return new_samples

    def _create_db_order(self, order: FastqOrder) -> Order:
        """Return an Order database object."""
        ticket_id: int = order._generated_ticket_id
        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=order.customer
        )
        return self.status_db.add_order(customer=customer, ticket_id=ticket_id)

    def _create_db_case_for_sample(
        self,
        sample: FastqSample,
        customer: Customer,
        order: FastqOrder,
    ) -> Case:
        """Return a Case database object."""
        ticket_id = str(order._generated_ticket_id)
        case_name: str = f"{sample.name}-{ticket_id}"
        priority: str = sample.priority
        case: Case = self.status_db.add_case(
            data_analysis=ORDER_TYPE_WORKFLOW_MAP[order.order_type],
            data_delivery=DataDelivery(order.delivery_type),
            name=case_name,
            priority=priority,
            ticket=ticket_id,
        )
        case.customer = customer
        return case

    def _create_db_sample(
        self, sample: FastqSample, order_name: str, customer: Customer, ticket_id: str
    ) -> Sample:
        """Return a Sample database object."""
        application_version: ApplicationVersion = (
            self.status_db.get_current_application_version_by_tag(tag=sample.application)
        )
        return self.status_db.add_sample(
            name=sample.name,
            sex=sample.sex or SexEnum.unknown,
            comment=sample.comment,
            internal_id=sample._generated_lims_id,
            ordered=datetime.now(),
            original_ticket=ticket_id,
            priority=sample.priority,
            tumour=sample.tumour,
            capture_kit=sample.capture_kit,
            subject_id=sample.subject_id,
            customer=customer,
            application_version=application_version,
            order=order_name,
        )

    def _create_maf_case(self, db_sample: Sample, db_order: Order, db_case: Case) -> None:
        """
        Add a MAF case and a relationship with the given sample to the current Status database
        transaction. This is done only if the given sample is non-tumour and  WGS.
        This function does not commit to the database.
        """
        if (
            not db_sample.is_tumour
            and db_sample.prep_category == SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING
        ):
            maf_order: Order = self.status_db.get_order_by_id(MAF_ORDER_ID)
            maf_case: Case = self.status_db.add_case(
                comment=f"MAF case for {db_case.internal_id} original order id {db_order.id}",
                data_analysis=Workflow.MIP_DNA,
                data_delivery=DataDelivery.NO_DELIVERY,
                name="_".join([db_sample.name, "MAF"]),
                panels=[GenePanelMasterList.OMIM_AUTO],
                priority=Priority.research,
                ticket=db_sample.original_ticket,
            )
            maf_case.customer = self.status_db.get_customer_by_internal_id(
                customer_internal_id=CustomerId.CG_INTERNAL_CUSTOMER
            )
            maf_case_sample: CaseSample = self.status_db.relate_sample(
                case=maf_case,
                sample=db_sample,
                status=StatusEnum.unknown,
                should_deliver_sample=False,
            )
            maf_order.cases.append(maf_case)
            self.status_db.add_multiple_items_to_store([maf_case, maf_case_sample])



from cg.services.orders.validation.errors.order_errors import OrderError


class CaseError(OrderError):
    case_index: int


class RepeatedCaseNameError(CaseError):
    field: str = "name"
    message: str = "Case name already used"


class InvalidGenePanelsError(CaseError):
    def __init__(self, case_index: int, panels: list[str]):
        message = "Invalid panels: " + ",".join(panels)
        super(CaseError, self).__init__(field="panels", case_index=case_index, message=message)


class RepeatedGenePanelsError(CaseError):
    field: str = "panels"
    message: str = "Gene panels must be unique"


class CaseNameNotAvailableError(CaseError):
    field: str = "name"
    message: str = "Case name already used in a previous order"


class CaseDoesNotExistError(CaseError):
    field: str = "internal_id"
    message: str = "The case does not exist"


class CaseOutsideOfCollaborationError(CaseError):
    field: str = "internal_id"
    message: str = "Case does not belong to collaboration"


class MultipleSamplesInCaseError(CaseError):
    field: str = "sample_errors"
    message: str = "Multiple samples in the same case not allowed"


class MoreThanTwoSamplesInCaseError(CaseError):
    field: str = "sample_errors"
    message: str = "More than two samples in the same case not allowed"


class NumberOfNormalSamplesError(CaseError):
    field: str = "sample_errors"


class DoubleNormalError(NumberOfNormalSamplesError):
    message: str = "Only one non-tumour sample is allowed per case"


class DoubleTumourError(NumberOfNormalSamplesError):
    message: str = "Only one tumour sample is allowed per case"


class NormalOnlyWGSError(NumberOfNormalSamplesError):
    message: str = "It is not possible to run the analysis on only one normal WGS sample."


class NewCaseWithoutAffectedSampleError(CaseError):
    field: str = "sample_errors"
    message: str = "Each case needs at least one affected sample"


class ExistingCaseWithoutAffectedSampleError(CaseError):
    field: str = "sample_errors"
    message: str = (
        "This case contains no affected sample. Please create a new case with at least one affected sample."
    )


class MultiplePrepCategoriesError(CaseError):
    field: str = "sample_errors"
    message: str = "Case cannot contain samples with incompatible applications"


class SamplesNotRelatedError(CaseError):
    field: str = "sample_errors"
    message: str



"""Backup related CLI commands."""

import logging
from pathlib import Path

import rich_click as click

from cg.apps.slurm.slurm_api import SlurmAPI
from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.constants.cli_options import DRY_RUN
from cg.constants.constants import SequencingRunDataAvailability
from cg.exc import (
    DsmcAlreadyRunningError,
    FlowCellError,
    IlluminaRunAlreadyBackedUpError,
    IlluminaRunEncryptionError,
    PdcError,
)
from cg.meta.encryption.encryption import EncryptionAPI
from cg.meta.tar.tar import TarAPI
from cg.models.cg_config import CGConfig
from cg.models.run_devices.illumina_run_directory_data import (
    IlluminaRunDirectoryData,
    get_sequencing_runs_from_path,
)
from cg.services.illumina.backup.backup_service import IlluminaBackupService
from cg.services.illumina.backup.encrypt_service import IlluminaRunEncryptionService
from cg.store.exc import EntryNotFoundError
from cg.store.models import IlluminaSequencingRun
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@click.group(context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_obj
def backup(context: CGConfig):
    """Backup utilities"""
    pass


@backup.command("illumina-runs")
@DRY_RUN
@click.pass_obj
def backup_illumina_runs(context: CGConfig, dry_run: bool):
    """Back-up Illumina runs."""
    pdc_service = context.pdc_service
    pdc_service.dry_run = dry_run
    encryption_api = EncryptionAPI(binary_path=context.encryption.binary_path, dry_run=dry_run)
    tar_api = TarAPI(binary_path=context.tar.binary_path, dry_run=dry_run)
    backup_service = IlluminaBackupService(
        encryption_api=encryption_api,
        pdc_archiving_directory=context.illumina_backup_service.pdc_archiving_directory,
        status_db=context.status_db,
        tar_api=tar_api,
        pdc_service=pdc_service,
        sequencing_runs_dir=context.run_instruments.illumina.sequencing_runs_dir,
        dry_run=dry_run,
    )
    backup_service.dry_run = dry_run
    status_db: Store = context.status_db
    runs_dir_data: list[IlluminaRunDirectoryData] = get_sequencing_runs_from_path(
        sequencing_run_dir=Path(context.run_instruments.illumina.sequencing_runs_dir)
    )
    for run_dir_data in runs_dir_data:
        try:
            sequencing_run: IlluminaSequencingRun = (
                status_db.get_illumina_sequencing_run_by_device_internal_id(run_dir_data.id)
            )
            backup_service.start_run_backup(
                run_dir_data=run_dir_data,
                sequencing_run=sequencing_run,
                status_db=status_db,
                binary_path=context.encryption.binary_path,
                encryption_dir=Path(context.encryption.encryption_dir),
                pigz_binary_path=context.pigz.binary_path,
                sbatch_parameter=context.illumina_backup_service.slurm_flow_cell_encryption.dict(),
            )
        except EntryNotFoundError as error:
            logging.error(f"{error}")
            continue
        except (
            DsmcAlreadyRunningError,
            IlluminaRunAlreadyBackedUpError,
            IlluminaRunEncryptionError,
            PdcError,
        ) as error:
            logging.error(f"{error}")


@backup.command("encrypt-illumina-runs")
@DRY_RUN
@click.pass_obj
def encrypt_illumina_runs(context: CGConfig, dry_run: bool):
    """Encrypt illumina runs."""
    status_db: Store = context.status_db
    runs: list[IlluminaRunDirectoryData] = get_sequencing_runs_from_path(
        sequencing_run_dir=Path(context.run_instruments.illumina.sequencing_runs_dir)
    )
    for run in runs:
        try:
            sequencing_run: IlluminaSequencingRun = (
                status_db.get_illumina_sequencing_run_by_device_internal_id(run.id)
            )
        except EntryNotFoundError as error:
            LOG.error(f"{error}")
            continue
        if sequencing_run.has_backup:
            LOG.debug(f"Run: {run.id} is already backed-up")
            continue
        illumina_run_encryption_service = IlluminaRunEncryptionService(
            binary_path=context.encryption.binary_path,
            dry_run=dry_run,
            encryption_dir=Path(context.encryption.encryption_dir),
            run_dir_data=run,
            pigz_binary_path=context.pigz.binary_path,
            slurm_api=SlurmAPI(),
            sbatch_parameter=context.illumina_backup_service.slurm_flow_cell_encryption.dict(),
            tar_api=TarAPI(binary_path=context.tar.binary_path, dry_run=dry_run),
        )
        try:
            illumina_run_encryption_service.start_encryption()
        except (FlowCellError, IlluminaRunEncryptionError) as error:
            logging.error(f"{error}")


@backup.command("fetch-illumina-run")
@click.option("-f", "--flow-cell-id", help="Retrieve a specific flow cell, ex. 'HCK2KDSXX'")
@DRY_RUN
@click.pass_obj
def fetch_illumina_run(context: CGConfig, dry_run: bool, flow_cell_id: str | None = None):
    """Fetch the first Illumina run in the requested queue from backup."""

    pdc_service = context.pdc_service
    pdc_service.dry_run = dry_run
    encryption_api = EncryptionAPI(binary_path=context.encryption.binary_path, dry_run=dry_run)
    tar_api = TarAPI(binary_path=context.tar.binary_path, dry_run=dry_run)
    context.meta_apis["backup_api"] = IlluminaBackupService(
        encryption_api=encryption_api,
        pdc_archiving_directory=context.illumina_backup_service.pdc_archiving_directory,
        status_db=context.status_db,
        tar_api=tar_api,
        pdc_service=pdc_service,
        sequencing_runs_dir=context.run_instruments.illumina.sequencing_runs_dir,
        dry_run=dry_run,
    )
    backup_api: IlluminaBackupService = context.meta_apis["backup_api"]

    status_db: Store = context.status_db
    sequencing_run: IlluminaSequencingRun | None = None
    if not flow_cell_id:
        LOG.info("Fetching first sequencing run in queue")
    try:
        if flow_cell_id:
            sequencing_run: IlluminaSequencingRun = (
                status_db.get_illumina_sequencing_run_by_device_internal_id(flow_cell_id)
            )
    except EntryNotFoundError as error:
        LOG.error(f"{error}")
        raise click.Abort

    retrieval_time: float | None = backup_api.fetch_sequencing_run(sequencing_run)

    if retrieval_time:
        hours = retrieval_time / 60 / 60
        LOG.info(f"Retrieval time: {hours:.1}h")
        return

    if not dry_run and sequencing_run:
        LOG.info(
            f"{sequencing_run}: updating sequencing run data availability to {SequencingRunDataAvailability.REQUESTED}"
        )
        status_db.update_illumina_sequencing_run_data_availability(
            sequencing_run=sequencing_run, data_availability=SequencingRunDataAvailability.REQUESTED
        )



from pathlib import Path

from pydantic import BaseModel, field_validator

from cg.constants.constants import GenomeVersion
from cg.constants.sample_sources import SourceType
from cg.utils.utils import replace_non_alphanumeric


class WorkflowParameters(BaseModel):
    input: Path
    outdir: Path


class NalloParameters(WorkflowParameters):
    """Model for Nallo parameters."""

    filter_variants_hgnc_ids: str


class RarediseaseParameters(WorkflowParameters):
    """Model for Raredisease parameters."""

    target_bed_file: str
    analysis_type: str
    save_mapped_as_cram: bool
    vcfanno_extra_resources: str
    vep_filters_scout_fmt: str
    sample_id_map: Path


class RNAFusionParameters(WorkflowParameters):
    """RNAFUSION parameters."""


class TaxprofilerParameters(WorkflowParameters):
    """Taxprofiler parameters."""


class TomteParameters(WorkflowParameters):
    """Model for Tomte parameters."""

    gene_panel_clinical_filter: Path
    tissue: str
    genome: str = GenomeVersion.HG38

    @field_validator("tissue", mode="before")
    @classmethod
    def restrict_tissue_values(cls, tissue: str | None) -> str:
        if tissue:
            return replace_non_alphanumeric(string=tissue)
        else:
            return SourceType.UNKNOWN

    @field_validator("genome", mode="before")
    @classmethod
    def restrict_genome_values(cls, genome: str) -> str:
        if genome == GenomeVersion.HG38:
            return GenomeVersion.GRCh38.value
        elif genome == GenomeVersion.HG19:
            return GenomeVersion.GRCh37.value



from typing import Callable

from cg.services.orders.validation.rules.case.rules import (
    validate_case_internal_ids_exist,
    validate_case_names_available,
    validate_case_names_not_repeated,
    validate_existing_cases_belong_to_collaboration,
)
from cg.services.orders.validation.rules.case_sample.rules import (
    validate_application_compatibility,
    validate_application_exists,
    validate_application_not_archived,
    validate_buffer_required,
    validate_buffer_skip_rc_condition,
    validate_concentration_interval_if_skip_rc,
    validate_concentration_required_if_skip_rc,
    validate_container_name_required,
    validate_existing_samples_belong_to_collaboration,
    validate_existing_samples_compatible_with_order_type,
    validate_sample_names_different_from_case_names,
    validate_sample_names_not_repeated,
    validate_samples_exist,
    validate_source_comment_required,
    validate_subject_ids_different_from_case_names,
    validate_subject_ids_different_from_sample_names,
    validate_subject_sex_consistency,
    validate_tube_container_name_unique,
    validate_volume_interval,
    validate_well_position_format,
    validate_well_positions_required,
    validate_wells_contain_at_most_one_sample,
)

MIP_RNA_CASE_RULES: list[Callable] = [
    validate_case_internal_ids_exist,
    validate_case_names_available,
    validate_case_names_not_repeated,
    validate_existing_cases_belong_to_collaboration,
]

MIP_RNA_CASE_SAMPLE_RULES: list[Callable] = [
    validate_application_compatibility,
    validate_application_exists,
    validate_application_not_archived,
    validate_buffer_required,
    validate_buffer_skip_rc_condition,
    validate_concentration_interval_if_skip_rc,
    validate_concentration_required_if_skip_rc,
    validate_container_name_required,
    validate_existing_samples_compatible_with_order_type,
    validate_existing_samples_belong_to_collaboration,
    validate_sample_names_different_from_case_names,
    validate_sample_names_not_repeated,
    validate_samples_exist,
    validate_source_comment_required,
    validate_subject_ids_different_from_case_names,
    validate_subject_ids_different_from_sample_names,
    validate_subject_sex_consistency,
    validate_tube_container_name_unique,
    validate_volume_interval,
    validate_well_position_format,
    validate_well_positions_required,
    validate_wells_contain_at_most_one_sample,
]



"""Code to handle files regarding compression and decompression."""

import datetime
import logging
from pathlib import Path

from housekeeper.store.models import File, Version

from cg.constants import HK_FASTQ_TAGS, SequencingFileTag
from cg.constants.compression import (
    FASTQ_DATETIME_DELTA,
    FASTQ_FIRST_READ_SUFFIX,
    FASTQ_SECOND_READ_SUFFIX,
)
from cg.models.compression_data import CompressionData

LOG = logging.getLogger(__name__)

# Functions to get common files


def get_hk_files_dict(tags: list[str], version_obj: Version) -> dict[Path, File]:
    """Fetch files from a version in Housekeeper."""
    hk_file: dict[Path, File] = {}
    tags: set = set(tags)
    for version_file in version_obj.files:
        file_tags: set[str] = {tag.name for tag in version_file.tags}
        if not file_tags.intersection(tags):
            continue
        LOG.debug(f"Found file {version_file.path}")
        path_obj: Path = Path(version_file.full_path)
        hk_file[path_obj] = version_file
    return hk_file


def is_file_in_version(version_obj: Version, path: Path) -> bool:
    """Check if a file is in a certain version considering relative path in the database."""
    return any(version_file.path in path.as_posix() for version_file in version_obj.files)


# Functions to get FASTQ like files


def get_spring_paths(version_obj: Version) -> list[CompressionData]:
    """Get all SPRING paths for a sample."""
    hk_files_dict: dict[Path, File] = get_hk_files_dict(
        tags=[SequencingFileTag.SPRING], version_obj=version_obj
    )
    spring_paths: list = []

    if hk_files_dict is None:
        return spring_paths

    spring_paths.extend(CompressionData(file_path.with_suffix("")) for file_path in hk_files_dict)
    return spring_paths


def get_fastq_stub(fastq_path: Path) -> Path:
    """Take a FASTQ file and return the stub (unique part).

    Example:
        fastq_file = /home/fastq_files/A_sequencing_run_R1_001.fastq.gz
        file_prefix = /home/fastq_files/A_sequencing_run
    """
    fastq_string: str = str(fastq_path)
    if FASTQ_FIRST_READ_SUFFIX in fastq_string:
        return Path(fastq_string.replace(FASTQ_FIRST_READ_SUFFIX, ""))
    if FASTQ_SECOND_READ_SUFFIX in fastq_string:
        return Path(fastq_string.replace(FASTQ_SECOND_READ_SUFFIX, ""))
    return None


def get_compression_data(fastq_files: list[Path]) -> list[CompressionData]:
    """Return a list of compression data objects.

    Each object has information about a pair of FASTQ files from the same run.
    """
    fastq_runs: set = set()
    compressions: list[CompressionData] = []
    for fastq_file in fastq_files:
        # file prefix is the run name identifier
        file_prefix: Path | None = get_fastq_stub(fastq_file)
        if file_prefix is None:
            LOG.info(f"Invalid FASTQ name {fastq_file}")
            continue
        run_name: str = str(file_prefix)
        if run_name not in fastq_runs:
            fastq_runs.add(run_name)
            compressions.append(CompressionData(file_prefix))
    return compressions


def get_fastq_files(sample_id: str, version_obj: Version) -> dict[str, dict]:
    """Get FASTQ files for sample."""
    hk_files_dict = get_hk_files_dict(tags=HK_FASTQ_TAGS, version_obj=version_obj)
    if hk_files_dict is None:
        return None

    fastq_dict = {}
    compression_objects: list[CompressionData] = get_compression_data(list(hk_files_dict.keys()))
    if not compression_objects:
        LOG.info(f"Could not find FASTQ files for {sample_id}")
        return None

    for compression_obj in compression_objects:
        if not check_fastqs(compression_obj):
            LOG.info(f"Skipping run {compression_obj.run_name}")
            continue
        fastq_dict[compression_obj.run_name] = {
            "compression_data": compression_obj,
            "hk_first": hk_files_dict[compression_obj.fastq_first],
            "hk_second": hk_files_dict[compression_obj.fastq_second],
        }

    return fastq_dict


def check_fastqs(compression_obj: CompressionData) -> bool:
    """Check if FASTQ files has the correct status.

    More specific this means to check
        - Did we get the full path of the file?
        - Does the file exist?
        - Do we have permissions?
        - Is the file actually a symlink?
        - Is the file a hardlink?
        - Is the file older than the specified time delta?
    """
    if not (
        compression_obj.is_absolute(compression_obj.fastq_first)
        or compression_obj.is_absolute(compression_obj.fastq_second)
    ):
        return False

    if not compression_obj.pair_exists():
        return False

    date_changed = compression_obj.get_change_date(compression_obj.fastq_first)
    today = datetime.datetime.now()

    # Check if date is older than FASTQ_DELTA
    if date_changed > today - FASTQ_DATETIME_DELTA:
        LOG.info("FASTQ files are not old enough")
        return False
    return True



from cg.services.orders.validation.errors.case_sample_errors import PedigreeError
from cg.services.orders.validation.order_types.mip_dna.models.case import MIPDNACase
from cg.services.orders.validation.order_types.raredisease.models.case import RarediseaseCase
from cg.services.orders.validation.order_types.tomte.models.case import TomteCase
from cg.services.orders.validation.rules.case_sample.pedigree.models import FamilyTree
from cg.services.orders.validation.rules.case_sample.pedigree.utils import validate_tree
from cg.store.store import Store


def get_pedigree_errors(
    case: TomteCase | MIPDNACase | RarediseaseCase, case_index: int, store: Store
) -> list[PedigreeError]:
    """Return a list of errors if any sample is labelled as its own ancestor in the family tree."""
    pedigree = FamilyTree(case=case, case_index=case_index, store=store)
    return validate_tree(pedigree)



import logging

from cg.exc import OrderError as OrderValidationError
from cg.models.orders.constants import OrderType
from cg.services.orders.validation.errors.case_errors import CaseError
from cg.services.orders.validation.errors.case_sample_errors import CaseSampleError
from cg.services.orders.validation.errors.order_errors import OrderError
from cg.services.orders.validation.errors.sample_errors import SampleError
from cg.services.orders.validation.errors.validation_errors import ValidationErrors
from cg.services.orders.validation.model_validator.model_validator import ModelValidator
from cg.services.orders.validation.models.order import Order
from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_type_maps import (
    ORDER_TYPE_MODEL_MAP,
    ORDER_TYPE_RULE_SET_MAP,
    RuleSet,
)
from cg.services.orders.validation.response_mapper import create_order_validation_response
from cg.services.orders.validation.utils import (
    apply_case_sample_validation,
    apply_case_validation,
    apply_order_validation,
    apply_sample_validation,
)
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class OrderValidationService:
    def __init__(self, store: Store):
        self.store = store

    def get_validation_response(self, raw_order: dict, order_type: OrderType, user_id: int) -> dict:
        model = ORDER_TYPE_MODEL_MAP[order_type]
        rule_set = ORDER_TYPE_RULE_SET_MAP[order_type]
        errors: ValidationErrors = self._get_errors(
            raw_order=raw_order, model=model, rule_set=rule_set, user_id=user_id
        )
        return create_order_validation_response(raw_order=raw_order, errors=errors)

    def parse_and_validate(self, raw_order: dict, order_type: OrderType, user_id: int) -> Order:
        model = ORDER_TYPE_MODEL_MAP[order_type]
        rule_set = ORDER_TYPE_RULE_SET_MAP[order_type]
        parsed_order, errors = ModelValidator.validate(order=raw_order, model=model)
        if parsed_order:
            parsed_order._user_id = user_id
            errors: ValidationErrors = self._get_rule_validation_errors(
                order=parsed_order,
                rule_set=rule_set,
            )
        if not errors.is_empty:
            LOG.error(errors.get_error_message())
            raise OrderValidationError(message="Order contained errors")
        return parsed_order

    def _get_errors(
        self, raw_order: dict, model: type[Order], rule_set: RuleSet, user_id: int
    ) -> ValidationErrors:
        parsed_order, errors = ModelValidator.validate(order=raw_order, model=model)
        if parsed_order:
            parsed_order._user_id = user_id
            errors: ValidationErrors = self._get_rule_validation_errors(
                order=parsed_order, rule_set=rule_set
            )
        return errors

    def _get_rule_validation_errors(self, order: Order, rule_set: RuleSet) -> ValidationErrors:

        case_errors = []
        case_sample_errors = []
        order_errors: list[OrderError] = apply_order_validation(
            rules=rule_set.order_rules,
            order=order,
            store=self.store,
        )
        sample_errors = []
        if isinstance(order, OrderWithCases):
            case_errors: list[CaseError] = apply_case_validation(
                rules=rule_set.case_rules,
                order=order,
                store=self.store,
            )
            case_sample_errors: list[CaseSampleError] = apply_case_sample_validation(
                rules=rule_set.case_sample_rules,
                order=order,
                store=self.store,
            )
        else:
            sample_errors: list[SampleError] = apply_sample_validation(
                rules=rule_set.sample_rules,
                order=order,
                store=self.store,
            )

        return ValidationErrors(
            case_errors=case_errors,
            case_sample_errors=case_sample_errors,
            order_errors=order_errors,
            sample_errors=sample_errors,
        )



from abc import abstractmethod
from pathlib import Path
from typing import Protocol, runtime_checkable


@runtime_checkable
class SampleSheetCreator(Protocol):
    @abstractmethod
    def create(self, case_id: str, file_path: Path) -> None:
        raise NotImplementedError(
            "Please implement create to conform to the SampleSheetFileCreator protocol"
        )



import logging
from datetime import datetime
from pathlib import Path
from typing import Any, Iterator, NamedTuple, Type

from dateutil.parser import parse
from pydantic import TypeAdapter
from pydantic.v1 import ValidationError

from cg.constants import Workflow
from cg.constants.constants import FileExtensions, FileFormat, GenomeVersion, MultiQC
from cg.constants.tb import AnalysisStatus
from cg.exc import CgError, HousekeeperStoreError, MetricsQCError
from cg.io.controller import ReadFile, WriteFile
from cg.io.json import read_json
from cg.meta.workflow.analysis import AnalysisAPI
from cg.models.analysis import NextflowAnalysis
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import (
    MetricsBase,
    MetricsDeliverablesCondition,
    MultiqcDataJson,
)
from cg.models.nf_analysis import FileDeliverable, WorkflowDeliverables
from cg.models.qc_metrics import QCMetrics
from cg.store.models import Analysis, Case, Sample

LOG = logging.getLogger(__name__)


class MultiQCSearchPattern(NamedTuple):
    pattern: str
    sample_id: str


class NfAnalysisAPI(AnalysisAPI):
    """Parent class for handling NF-core analyses."""

    def __init__(self, config: CGConfig, workflow: Workflow):
        super().__init__(workflow=workflow, config=config)
        self.workflow: Workflow = workflow
        self.root_dir: str | None = None
        self.workflow_bin_path: str | None = None
        self.references: str | None = None
        self.profile: str | None = None
        self.conda_env: str | None = None
        self.conda_binary: str | None = None
        self.platform: str | None = None
        self.params: str | None = None
        self.workflow_config_path: str | None = None
        self.resources: str | None = None
        self.tower_binary_path: str | None = None
        self.tower_workflow: str | None = None
        self.account: str | None = None
        self.email: str | None = None
        self.revision: str | None = None

    @property
    def root(self) -> str:
        return self.root_dir

    @property
    def is_multiqc_pattern_search_exact(self) -> bool:
        """Return True if only exact pattern search is allowed to collect metrics information from MultiQC file.
        If false, pattern must be present but does not need to be exact."""
        return False

    def get_case_path(self, case_id: str) -> Path:
        """Path to case working directory."""
        return Path(self.root_dir, case_id)

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """Path to deliverables file for a case."""
        return Path(self.get_case_path(case_id), f"{case_id}_deliverables").with_suffix(
            FileExtensions.YAML
        )

    def get_metrics_deliverables_path(self, case_id: str) -> Path:
        """Return a path where the <case>_metrics_deliverables.yaml file should be located."""
        return Path(self.root_dir, case_id, f"{case_id}_metrics_deliverables").with_suffix(
            FileExtensions.YAML
        )

    def verify_deliverables_file_exists(self, case_id: str) -> None:
        """Raise an error if a deliverable file is not found."""
        if not Path(self.get_deliverables_file_path(case_id=case_id)).exists():
            raise CgError(f"No deliverables file found for case {case_id}")

    @staticmethod
    def write_deliverables_file(
        deliverables_content: dict, file_path: Path, file_format=FileFormat.YAML
    ) -> None:
        """Write deliverables file."""
        WriteFile.write_file_from_content(
            content=deliverables_content, file_format=file_format, file_path=file_path
        )

    def get_deliverables_template_content(self) -> list[dict[str, str]]:
        """Return deliverables file template content."""
        LOG.debug("Getting deliverables file template content")
        return ReadFile.get_content_from_file(
            file_format=FileFormat.YAML,
            file_path=self.get_bundle_filenames_path(),
        )

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return bundle filenames path."""
        raise NotImplementedError

    @staticmethod
    def get_formatted_file_deliverable(
        file_template: dict[str | None, str | None],
        case_id: str,
        sample_id: str,
        sample_name: str,
        case_path: str,
    ) -> FileDeliverable:
        """Return the formatted file deliverable with the case and sample attributes."""
        deliverables = file_template.copy()
        for deliverable_field, deliverable_value in file_template.items():
            if deliverable_value is None:
                continue
            deliverables[deliverable_field] = (
                deliverables[deliverable_field]
                .replace("CASEID", case_id)
                .replace("SAMPLEID", sample_id)
                .replace("SAMPLENAME", sample_name)
                .replace("PATHTOCASE", case_path)
            )
            LOG.debug(deliverables[deliverable_field])
        return FileDeliverable(**deliverables)

    def get_deliverables_for_sample(
        self, sample: Sample, case_id: str, template: list[dict[str, str]]
    ) -> list[FileDeliverable]:
        """Return a list of FileDeliverables for each sample."""
        sample_id: str = sample.internal_id
        sample_name: str = sample.name
        case_path = str(self.get_case_path(case_id=case_id))
        files: list[FileDeliverable] = []
        for file in template:
            files.append(
                self.get_formatted_file_deliverable(
                    file_template=file,
                    case_id=case_id,
                    sample_id=sample_id,
                    sample_name=sample_name,
                    case_path=case_path,
                )
            )
        return files

    def get_deliverables_for_case(self, case_id: str) -> WorkflowDeliverables:
        """Return workflow deliverables for a given case."""
        deliverable_template: list[dict] = self.get_deliverables_template_content()
        samples: list[Sample] = self.status_db.get_samples_by_case_id(case_id=case_id)
        files: list[FileDeliverable] = []

        for sample in samples:
            bundles_per_sample = self.get_deliverables_for_sample(
                sample=sample, case_id=case_id, template=deliverable_template
            )
            files.extend(bundle for bundle in bundles_per_sample if bundle not in files)
        return WorkflowDeliverables(files=files)

    def get_multiqc_json_path(self, case_id: str) -> Path:
        """Return the path of the multiqc_data.json file."""
        return Path(
            self.root_dir,
            case_id,
            MultiQC.MULTIQC,
            MultiQC.MULTIQC_DATA,
            MultiQC.MULTIQC_DATA + FileExtensions.JSON,
        )

    def get_qc_conditions_for_workflow(self, sample_id: str) -> dict:
        """Get nf-core workflow metrics constants."""
        return {}

    def get_multiqc_search_patterns(self, case_id: str) -> list[MultiQCSearchPattern]:
        """Return a list of MultiQC search patterns for a case.
        Each pattern pairs a search string with its corresponding sample ID.
        Search patterns are typically sample IDs used to extract metrics from MultiQC data."""
        sample_ids: Iterator[str] = self.status_db.get_sample_ids_by_case_id(case_id)
        return [
            MultiQCSearchPattern(pattern=sample_id, sample_id=sample_id) for sample_id in sample_ids
        ]

    @staticmethod
    def get_deduplicated_metrics(metrics: list[MetricsBase]) -> list[MetricsBase]:
        """Return deduplicated metrics based on metric ID and name. If duplicated entries are found
        only the first one will be kept."""
        deduplicated_metric_id_name = set([])
        deduplicated_metrics: list = []
        for metric in metrics:
            if (metric.id, metric.name) not in deduplicated_metric_id_name:
                deduplicated_metric_id_name.add((metric.id, metric.name))
                deduplicated_metrics.append(metric)
        return deduplicated_metrics

    def get_multiqc_data_json(self, case_id: str) -> MultiqcDataJson:
        return MultiqcDataJson(**read_json(file_path=self.get_multiqc_json_path(case_id=case_id)))

    def get_multiqc_json_metrics(self, case_id: str) -> list[MetricsBase]:
        """Return a list of the metrics specified in a MultiQC json file."""
        multiqc_json: MultiqcDataJson = self.get_multiqc_data_json(case_id=case_id)
        metrics = []
        for pattern in self.get_multiqc_search_patterns(case_id=case_id):
            metrics_for_pattern: list[MetricsBase] = self.get_multiqc_metrics_for_sample(
                search_pattern=pattern.pattern,
                multiqc_json=multiqc_json,
                sample_id=pattern.sample_id,
                exact_match=self.is_multiqc_pattern_search_exact,
            )
            metrics.extend(metrics_for_pattern)
        metrics = self.get_deduplicated_metrics(metrics=metrics)
        return metrics

    @staticmethod
    def _is_pattern_found(pattern: str, text: str, exact_match: bool) -> bool:
        if exact_match:
            is_pattern_found: bool = pattern == text
        else:
            is_pattern_found: bool = pattern in text
        return is_pattern_found

    def get_multiqc_metrics_for_sample(
        self,
        search_pattern: str,
        multiqc_json: MultiqcDataJson,
        sample_id: str,
        exact_match: bool = False,
    ) -> list[MetricsBase]:
        """Parse a MultiqcDataJson and returns a list of metrics."""
        metrics: list[MetricsBase] = []
        list_of_metric_dicts: list[dict[str, Any]] = self._get_list_of_metric_dicts(multiqc_json)

        for section in list_of_metric_dicts:
            for subsection, metrics_dict in section.items():
                if self._is_pattern_found(
                    pattern=search_pattern, text=subsection, exact_match=exact_match
                ):
                    for metric_name, metric_value in metrics_dict.items():
                        metric: MetricsBase = self.get_multiqc_metric(
                            metric_name=metric_name, metric_value=metric_value, sample_id=sample_id
                        )
                        metrics.append(metric)
        return metrics

    def _get_list_of_metric_dicts(self, multiqc_json: MultiqcDataJson) -> list[dict[str, Any]]:
        if metric_dicts := multiqc_json.report_general_stats_data:
            return metric_dicts
        else:
            raise ValueError("No report_general_stats_data found in MultiqcDataJson")

    def get_multiqc_metric(
        self, metric_name: str, metric_value: str | int | float, sample_id: str
    ) -> MetricsBase:
        """Return a MetricsBase object for a given metric."""
        return MetricsBase(
            header=None,
            id=sample_id,
            input=MultiQC.MULTIQC_DATA + FileExtensions.JSON,
            name=metric_name,
            step=MultiQC.MULTIQC,
            value=metric_value,
            condition=self.get_qc_conditions_for_workflow(sample_id).get(metric_name, None),
        )

    @staticmethod
    def ensure_mandatory_metrics_present(metrics: list[MetricsBase]) -> None:
        return None

    def create_metrics_deliverables_content(self, case_id: str) -> dict[str, list[dict[str, Any]]]:
        """Create the content of metrics deliverables file."""
        metrics: list[MetricsBase] = self.get_multiqc_json_metrics(case_id=case_id)
        self.ensure_mandatory_metrics_present(metrics=metrics)
        return {"metrics": [metric.model_dump() for metric in metrics]}

    def write_metrics_deliverables(self, case_id: str, dry_run: bool = False) -> None:
        """Write <case>_metrics_deliverables.yaml file."""
        metrics_deliverables_path: Path = self.get_metrics_deliverables_path(case_id=case_id)
        content: dict = self.create_metrics_deliverables_content(case_id=case_id)
        if dry_run:
            LOG.info(
                f"Dry-run: metrics deliverables file would be written to {metrics_deliverables_path.as_posix()}"
            )
            return

        LOG.info(f"Writing metrics deliverables file to {metrics_deliverables_path.as_posix()}")
        WriteFile.write_file_from_content(
            content=content,
            file_format=FileFormat.YAML,
            file_path=metrics_deliverables_path,
        )

    def validate_qc_metrics(self, case_id: str, dry_run: bool = False) -> None:
        """Validate the information from a QC metrics deliverable file."""

        if dry_run:
            LOG.info("Dry-run: QC metrics validation would be performed")
            return

        LOG.info("Validating QC metrics")
        try:
            metrics_deliverables_path: Path = self.get_metrics_deliverables_path(case_id=case_id)
            qc_metrics_raw: dict = ReadFile.get_content_from_file(
                file_format=FileFormat.YAML, file_path=metrics_deliverables_path
            )
            MetricsDeliverablesCondition(**qc_metrics_raw)
        except MetricsQCError as error:
            LOG.error(f"QC metrics failed for {case_id}, with: {error}")
            self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.FAILED)
            samples = self.status_db.get_samples_by_case_id(case_id=case_id)
            for sample in samples:
                trailblazer_comment = str(error).replace(
                    f"{sample.internal_id} - ", f"{sample.name} ({sample.internal_id}) - "
                )
            self.trailblazer_api.add_comment(case_id=case_id, comment=trailblazer_comment)
            raise MetricsQCError from error
        except CgError as error:
            LOG.error(f"Could not create metrics deliverables file: {error}")
            self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.ERROR)
            raise CgError from error
        self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.COMPLETED)

    def metrics_deliver(self, case_id: str, dry_run: bool):
        """Create and validate a metrics deliverables file for given case id."""
        self.status_db.verify_case_exists(case_internal_id=case_id)
        self.write_metrics_deliverables(case_id=case_id, dry_run=dry_run)
        self.validate_qc_metrics(case_id=case_id, dry_run=dry_run)

    def report_deliver(self, case_id: str, dry_run: bool = False, force: bool = False) -> None:
        """Write deliverables file."""
        self.status_db.verify_case_exists(case_id)
        self.trailblazer_api.verify_latest_analysis_is_completed(case_id=case_id, force=force)
        if dry_run:
            LOG.info(f"Dry-run: Would have created delivery files for case {case_id}")
            return
        workflow_content: WorkflowDeliverables = self.get_deliverables_for_case(case_id=case_id)
        self.write_deliverables_file(
            deliverables_content=workflow_content.model_dump(),
            file_path=self.get_deliverables_file_path(case_id=case_id),
        )
        LOG.info(
            f"Writing deliverables file in {self.get_deliverables_file_path(case_id=case_id).as_posix()}"
        )

    def store_analysis_housekeeper(
        self, case_id: str, comment: str | None = None, dry_run: bool = False, force: bool = False
    ) -> None:
        """
        Store a finished Nextflow analysis in Housekeeper and StatusDB.

        Raises:
            HousekeeperStoreError: If the deliverables file is malformed or if the bundle could not be stored.
        """
        try:
            self.status_db.verify_case_exists(case_id)
            self.trailblazer_api.verify_latest_analysis_is_completed(case_id=case_id, force=force)
            self.verify_deliverables_file_exists(case_id)
            _, version = self.create_housekeeper_bundle(
                case_id=case_id, dry_run=dry_run, force=force
            )
            self.update_analysis_as_completed_statusdb(
                case_id=case_id,
                hk_version_id=version.id,
                comment=comment,
                dry_run=dry_run,
                force=force,
            )
            self.set_statusdb_action(case_id=case_id, action=None, dry_run=dry_run)
        except ValidationError as error:
            raise HousekeeperStoreError(f"Deliverables file is malformed: {error}")
        except Exception as error:
            self.housekeeper_api.rollback()
            self.status_db.session.rollback()
            raise HousekeeperStoreError(
                f"Could not store bundle in Housekeeper and StatusDB: {error}"
            )

    def store(
        self, case_id: str, comment: str | None = None, dry_run: bool = False, force: bool = False
    ):
        """Generate deliverable files for a case and store in Housekeeper if they
        pass QC metrics checks."""
        is_latest_analysis_qc: bool = self.trailblazer_api.is_latest_analysis_qc(case_id)
        is_latest_analysis_completed: bool = self.trailblazer_api.is_latest_analysis_completed(
            case_id
        )
        if not is_latest_analysis_qc and not is_latest_analysis_completed and not force:
            LOG.error(
                "Case not stored. Trailblazer status must be either QC or COMPLETED to be able to store"
            )
            raise ValueError

        if (
            is_latest_analysis_qc
            or not self.get_metrics_deliverables_path(case_id=case_id).exists()
            and not force
        ):
            LOG.info(f"Generating metrics file and performing QC checks for {case_id}")
            self.metrics_deliver(case_id=case_id, dry_run=dry_run)
        LOG.info(f"Storing analysis for {case_id}")
        self.report_deliver(case_id=case_id, dry_run=dry_run, force=force)
        self.store_analysis_housekeeper(
            case_id=case_id, comment=comment, dry_run=dry_run, force=force
        )

    def get_cases_to_store(self) -> list[Case]:
        """Return cases where analysis finished successfully,
        and is ready to be stored in Housekeeper."""
        return [
            case
            for case in self.status_db.get_running_cases_in_workflow(workflow=self.workflow)
            if self.trailblazer_api.is_latest_analysis_completed(case_id=case.internal_id)
            or self.trailblazer_api.is_latest_analysis_qc(case_id=case.internal_id)
        ]

    def get_genome_build(self, case_id: str) -> GenomeVersion:
        raise NotImplementedError

    def parse_analysis(
        self, qc_metrics_raw: list[MetricsBase], qc_metrics_model: Type[QCMetrics], **kwargs
    ) -> NextflowAnalysis:
        """Parse Nextflow output analysis files and return an analysis model."""
        sample_metrics: dict[str, dict] = {}
        for metric in qc_metrics_raw:
            try:
                sample_metrics[metric.id].update({metric.name.lower(): metric.value})
            except KeyError:
                sample_metrics[metric.id] = {metric.name.lower(): metric.value}
        pydantic_parser = TypeAdapter(dict[str, qc_metrics_model])
        return NextflowAnalysis(sample_metrics=pydantic_parser.validate_python(sample_metrics))

    def get_latest_metadata(self, case_id: str) -> NextflowAnalysis:
        """Return analysis output of a Nextflow case."""
        qc_metrics: list[MetricsBase] = self.get_multiqc_json_metrics(case_id)
        return self.parse_analysis(qc_metrics_raw=qc_metrics)

    def clean_past_run_dirs(self, before_date: str, skip_confirmation: bool = False) -> None:
        """Clean past run directories"""
        before_date: datetime = parse(before_date)
        analyses_to_clean: list[Analysis] = self.get_analyses_to_clean(before_date)
        LOG.info(f"Cleaning {len(analyses_to_clean)} analyses created before {before_date}")

        for analysis in analyses_to_clean:
            case_id = analysis.case.internal_id
            case_path = self.get_case_path(case_id)
            try:
                LOG.info(f"Cleaning output for {case_id}")
                self.clean_run_dir(
                    case_id=case_id, skip_confirmation=skip_confirmation, case_path=case_path
                )
            except FileNotFoundError:
                continue
            except Exception as error:
                LOG.error(f"Failed to clean directories for case {case_id} - {repr(error)}")
        LOG.info(f"Done cleaning {self.workflow} output")



"""Module for the sequencing time collector."""

from cg.constants.sequencing import Sequencers
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.file_parsing.sequencing_times.hiseq_2500_sequencing_times_service import (
    Hiseq2500SequencingTimesService,
)
from cg.services.illumina.file_parsing.sequencing_times.hiseq_x_sequencing_times_service import (
    HiseqXSequencingTimesService,
)
from cg.services.illumina.file_parsing.sequencing_times.novaseq_6000_sequencing_times import (
    Novaseq6000SequencingTimesService,
)
from cg.services.illumina.file_parsing.sequencing_times.novaseq_x_sequencing_times_service import (
    NovaseqXSequencingTimesService,
)
from cg.services.illumina.file_parsing.sequencing_times.sequencing_time_service import (
    SequencingTimesService,
)


class CollectSequencingTimes:
    """Class to collect sequencing times for Illumina sequencing runs."""

    @staticmethod
    def _get_sequencing_times_service(run_directory_data) -> SequencingTimesService:
        sequencer_times_services: dict = {
            Sequencers.NOVASEQX: NovaseqXSequencingTimesService,
            Sequencers.NOVASEQ: Novaseq6000SequencingTimesService,
            Sequencers.HISEQX: HiseqXSequencingTimesService,
            Sequencers.HISEQGA: Hiseq2500SequencingTimesService,
        }
        return sequencer_times_services[run_directory_data.sequencer_type]

    def get_end_time(self, run_directory_data: IlluminaRunDirectoryData):
        """Get the end time of the sequencing run."""
        times_service: SequencingTimesService = self._get_sequencing_times_service(
            run_directory_data
        )
        return times_service.get_end_time(run_directory_data=run_directory_data)

    def get_start_time(self, run_directory_data: IlluminaRunDirectoryData):
        """Get the start time of the sequencing run."""
        times_service: SequencingTimesService = self._get_sequencing_times_service(
            run_directory_data
        )
        return times_service.get_start_time(run_directory_data=run_directory_data)



"""
Module for mutacc-auto API
"""

import copy
import logging
import subprocess

from cg.constants.constants import FileFormat
from cg.io.controller import WriteStream

LOG = logging.getLogger(__name__)


class MutaccAutoAPI:
    """
    API for mutacc-auto
    """

    def __init__(self, config: dict):
        self.mutacc_auto_config = config["mutacc_auto"]["config_path"]
        self.mutacc_auto_binary = config["mutacc_auto"]["binary_path"]
        self.mutacc_padding = config["mutacc_auto"]["padding"]
        self.base_call = [self.mutacc_auto_binary, "--config-file", self.mutacc_auto_config]

    def extract_reads(self, case: dict, variants: dict):
        """
        Use mutacc-auto extract command to extract the relevant reads from
        the case

        Args:
            case (dict): case dictionary from scout
            variants (dict): variants dict from scout
            padding (int): padding (bp:s) around genomic region of variant

        """

        extract_call = copy.deepcopy(self.base_call)

        extract_call.extend(
            [
                "extract",
                "--variants",
                WriteStream.write_stream_from_content(
                    content=variants, file_format=FileFormat.JSON
                ),
                "--case",
                WriteStream.write_stream_from_content(content=case, file_format=FileFormat.JSON),
                "--padding",
                str(self.mutacc_padding),
                "--conda",
            ]
        )

        run_command(extract_call)

    def import_reads(self):
        """
        Upload the cases and extracted reads to mutacc DB
        """

        import_call = copy.deepcopy(self.base_call)
        import_call.extend(["import"])

        run_command(import_call)


def run_command(command: list):
    """
    Use subprocess.run to run command

    Args:
        command (list): list of commands
    """

    completed_process = subprocess.run(args=command, check=False)
    returncode = completed_process.returncode
    if returncode != 0:
        LOG.warning(f"process {' '.join(command)} ended with exitcode {returncode}")
        raise subprocess.CalledProcessError(returncode=returncode, cmd=command)



from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import (
    REMINDER_TO_DOWNLOAD_MESSAGE,
    get_caesar_delivery_path,
)
from cg.store.models import Case


class FastqMessage(DeliveryMessage):
    def create_message(self, cases: list[Case]) -> str:
        delivery_path: str = get_caesar_delivery_path(cases[0])
        sample_names: str = "\n".join([sample.name for case in cases for sample in case.samples])
        number_of_samples: int = sum(len(case.samples) for case in cases)
        sample_s_: str = "sample" if number_of_samples == 1 else "samples"
        return (
            f"Hello,\n\n"
            f"The fastq files for the following {sample_s_} have been uploaded to your inbox on Caesar:\n\n"
            f"{sample_names}\n\n"
            "Available under: \n\n"
            f"{delivery_path} \n\n"
            f"{REMINDER_TO_DOWNLOAD_MESSAGE}"
        )



from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.models.existing_case import ExistingCase
from cg.services.orders.validation.models.existing_sample import ExistingSample
from cg.services.orders.validation.order_types.balsamic.models.case import BalsamicCase
from cg.services.orders.validation.order_types.balsamic.models.sample import BalsamicSample
from cg.services.orders.validation.order_types.balsamic_umi.models.case import BalsamicUmiCase
from cg.services.orders.validation.order_types.balsamic_umi.models.sample import BalsamicUmiSample
from cg.services.orders.validation.order_types.mip_dna.models.case import MIPDNACase
from cg.services.orders.validation.order_types.mip_dna.models.sample import MIPDNASample
from cg.services.orders.validation.order_types.raredisease.models.case import RarediseaseCase
from cg.services.orders.validation.order_types.raredisease.models.sample import RarediseaseSample
from cg.store.models import Application
from cg.store.models import Case as DbCase
from cg.store.models import Customer, Sample
from cg.store.store import Store


def contains_duplicates(input_list: list) -> bool:
    return len(set(input_list)) != len(input_list)


def is_double_tumour(case: BalsamicCase | BalsamicUmiCase, store: Store) -> bool:
    return len(case.samples) == 2 and get_number_of_tumours(case=case, store=store) == 2


def is_double_normal(case: BalsamicCase | BalsamicUmiCase, store: Store) -> bool:
    return len(case.samples) == 2 and get_number_of_tumours(case=case, store=store) == 0


def get_number_of_tumours(case: BalsamicCase | BalsamicUmiCase, store: Store) -> int:
    number_of_tumours = 0
    for sample in case.samples:
        if sample.is_new and sample.tumour:
            number_of_tumours += 1
        elif not sample.is_new:
            db_sample: Sample = store.get_sample_by_internal_id(sample.internal_id)
            if db_sample.is_tumour:
                number_of_tumours += 1
    return number_of_tumours


def is_normal_only_wgs(case: BalsamicCase | BalsamicUmiCase, store: Store) -> bool:
    return len(case.samples) == 1 and _is_sample_wgs_normal(sample=case.samples[0], store=store)


def _is_sample_wgs_normal(
    sample: BalsamicSample | BalsamicUmiSample | ExistingSample, store: Store
) -> bool:
    if sample.is_new:
        application: Application = store.get_application_by_tag(sample.application)
        return (
            application.prep_category == SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING
            and not sample.tumour
        )
    else:
        db_sample: Sample = store.get_sample_by_internal_id(sample.internal_id)
        return (
            db_sample.prep_category == SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING
            and not db_sample.is_tumour
        )


def is_case_not_from_collaboration(case: ExistingCase, customer_id: str, store: Store) -> bool:
    db_case: DbCase | None = store.get_case_by_internal_id(case.internal_id)
    customer: Customer | None = store.get_customer_by_internal_id(customer_id)
    return db_case and customer and db_case.customer not in customer.collaborators


def is_sample_in_case(case: Case, sample_name: str, store: Store) -> bool:
    if case.get_new_sample(sample_name):
        return True
    elif case.get_existing_sample_from_db(sample_name=sample_name, store=store):
        return True
    return False


def get_case_prep_categories(case: Case, store: Store) -> set[str]:
    """
    Return a set with all prep categories of the samples in the case
    if the sample and its application exist, and has a prep category.
    """
    prep_categories: set[str] = set()
    for _, sample in case.enumerated_new_samples:
        application: Application | None = store.get_application_by_tag(sample.application)
        if application and application.prep_category:
            prep_categories.add(application.prep_category)

    for _, sample in case.enumerated_existing_samples:
        db_sample: Sample | None = store.get_sample_by_internal_id(sample.internal_id)
        if db_sample and db_sample.prep_category:
            prep_categories.add(db_sample.prep_category)
    return prep_categories


def does_case_exist(case: MIPDNACase | RarediseaseCase | ExistingCase, store: Store):
    if isinstance(case, ExistingCase):
        return bool(store.get_case_by_internal_id(case.internal_id))
    return True


def is_single_sample_case(case: MIPDNACase | RarediseaseCase | ExistingCase, store: Store):
    if isinstance(case, ExistingCase):
        db_case: DbCase = store.get_case_by_internal_id_strict(case.internal_id)
        contains_one_sample = bool(len(db_case.samples) == 1)
    else:
        contains_one_sample = bool(len(case.samples) == 1)
    return contains_one_sample


def is_sample_related_in_case(
    sample: MIPDNASample | RarediseaseSample | ExistingSample,
    case: MIPDNACase | RarediseaseCase,
    store: Store,
):
    if not (sample.mother or sample.father):
        sample_name: str = get_sample_name(sample=sample, store=store)
        if not any(
            sample_name in [sample_in_case.mother, sample_in_case.father]
            for sample_in_case in case.samples
        ):
            return False
    return True


def get_sample_name(sample: MIPDNASample | RarediseaseSample | ExistingSample, store: Store) -> str:
    if isinstance(sample, ExistingSample):
        sample_name: str = store.get_sample_by_internal_id_strict(sample.internal_id).name
    else:
        sample_name: str = sample.name
    return sample_name



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.pacbio_long_read.constants import PacbioDeliveryType
from cg.services.orders.validation.order_types.pacbio_long_read.models.sample import PacbioSample


class PacbioOrder(OrderWithSamples[PacbioSample]):
    delivery_type: PacbioDeliveryType



from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.raredisease.constants import RarediseaseDeliveryType
from cg.services.orders.validation.order_types.raredisease.models.case import RarediseaseCase
from cg.services.orders.validation.order_types.raredisease.models.sample import RarediseaseSample


class RarediseaseOrder(OrderWithCases[RarediseaseCase, RarediseaseSample]):
    delivery_type: RarediseaseDeliveryType



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.fastq.constants import FastqDeliveryType
from cg.services.orders.validation.order_types.fastq.models.sample import FastqSample


class FastqOrder(OrderWithSamples[FastqSample]):
    delivery_type: FastqDeliveryType



from cg.models.orders.constants import OrderType
from cg.services.orders.validation.errors.sample_errors import (
    ApplicationArchivedError,
    ApplicationNotCompatibleError,
    ApplicationNotValidError,
    BufferInvalidError,
    CaptureKitInvalidError,
    CaptureKitMissingError,
    ConcentrationInvalidIfSkipRCError,
    ConcentrationRequiredError,
    ContainerNameMissingError,
    ContainerNameRepeatedError,
    IndexNumberMissingError,
    IndexNumberOutOfRangeError,
    InvalidVolumeError,
    MissingSourceCommentError,
    OccupiedWellError,
    PoolApplicationError,
    PoolPriorityError,
    SampleError,
    SampleNameNotAvailableControlError,
    SampleNameRepeatedError,
    VolumeRequiredError,
    WellFormatError,
    WellFormatRmlError,
    WellPositionMissingError,
    WellPositionRmlMissingError,
)
from cg.services.orders.validation.models.order_aliases import (
    OrderWithControlSamples,
    OrderWithIndexedSamples,
)
from cg.services.orders.validation.models.sample_aliases import IndexedSample
from cg.services.orders.validation.order_types.fastq.models.order import FastqOrder
from cg.services.orders.validation.order_types.metagenome.models.order import MetagenomeOrder
from cg.services.orders.validation.order_types.microsalt.models.order import OrderWithSamples
from cg.services.orders.validation.order_types.pacbio_long_read.models.order import PacbioOrder
from cg.services.orders.validation.order_types.taxprofiler.models.order import TaxprofilerOrder
from cg.services.orders.validation.rules.sample.utils import (
    PlateSamplesValidator,
    get_indices_for_repeated_sample_names,
    get_indices_for_tube_repeated_container_name,
    get_sample_name_not_available_errors,
    has_multiple_applications,
    has_multiple_priorities,
    is_container_name_missing,
    is_index_number_missing,
    is_index_number_out_of_range,
    is_invalid_well_format,
    is_invalid_well_format_rml,
    validate_buffers_are_allowed,
    validate_concentration_interval,
    validate_concentration_required,
)
from cg.services.orders.validation.rules.utils import (
    is_application_compatible,
    is_invalid_capture_kit,
    is_sample_missing_capture_kit,
    is_volume_invalid,
    is_volume_missing,
)
from cg.store.store import Store


def validate_application_compatibility(
    order: OrderWithSamples,
    store: Store,
    **kwargs,
) -> list[ApplicationNotCompatibleError]:
    """
    Validate that the applications of all samples in the order are compatible with the order type.
    Applicable to all order types.
    """
    errors: list[ApplicationNotCompatibleError] = []
    order_type: OrderType = order.order_type
    for sample_index, sample in order.enumerated_samples:
        compatible: bool = is_application_compatible(
            order_type=order_type,
            application_tag=sample.application,
            store=store,
        )
        if not compatible:
            error = ApplicationNotCompatibleError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_application_exists(
    order: OrderWithSamples, store: Store, **kwargs
) -> list[ApplicationNotValidError]:
    """
    Validate that the applications of all samples in the order exist in the database.
    Applicable to all order types.
    """
    errors: list[ApplicationNotValidError] = []
    for sample_index, sample in order.enumerated_samples:
        if not store.get_application_by_tag(sample.application):
            error = ApplicationNotValidError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_applications_not_archived(
    order: OrderWithSamples, store: Store, **kwargs
) -> list[ApplicationArchivedError]:
    """
    Validate that none of the applications of the samples in the order are archived.
    Applicable to all order types.
    """
    errors: list[ApplicationArchivedError] = []
    for sample_index, sample in order.enumerated_samples:
        if store.is_application_archived(sample.application):
            error = ApplicationArchivedError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_buffer_skip_rc_condition(order: FastqOrder, **kwargs) -> list[BufferInvalidError]:
    """
    Validate that the sample buffers allow skipping reception control if that option is true.
    Only applicable to order types that have targeted sequencing applications (TGS).
    """
    errors: list[BufferInvalidError] = []
    if order.skip_reception_control:
        errors.extend(validate_buffers_are_allowed(order))
    return errors


def validate_capture_kit_compatible(
    order: FastqOrder, store: Store, **kwargs
) -> list[CaptureKitInvalidError]:
    """Validates that the capture kit is in our Bed table, and active."""
    errors: list[CaptureKitInvalidError] = []
    for sample_index, sample in order.enumerated_samples:
        if is_invalid_capture_kit(sample=sample, store=store):
            error = CaptureKitInvalidError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_capture_kit_required(
    order: FastqOrder, store: Store, **kwargs
) -> list[CaptureKitMissingError]:
    """Validates that capture kit is required for TGS samples"""
    errors: list[CaptureKitMissingError] = []
    for sample_index, sample in order.enumerated_samples:
        if is_sample_missing_capture_kit(sample=sample, store=store):
            error = CaptureKitMissingError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_concentration_interval_if_skip_rc(
    order: FastqOrder, store: Store, **kwargs
) -> list[ConcentrationInvalidIfSkipRCError]:
    """
    Validate that all samples have an allowed concentration if the order skips reception control.
    Only applicable to order types that have targeted sequencing applications (TGS).
    """
    errors: list[ConcentrationInvalidIfSkipRCError] = []
    if order.skip_reception_control:
        errors.extend(validate_concentration_interval(order=order, store=store))
    return errors


def validate_concentration_required_if_skip_rc(
    order: FastqOrder, **kwargs
) -> list[ConcentrationRequiredError]:
    """
    Validate that all samples have a concentration if the order skips reception control.
    Only applicable to order types that have targeted sequencing applications (TGS).
    """
    errors: list[ConcentrationRequiredError] = []
    if order.skip_reception_control:
        errors.extend(validate_concentration_required(order))
    return errors


def validate_container_name_required(
    order: OrderWithSamples, **kwargs
) -> list[ContainerNameMissingError]:
    """
    Validate that the container names are present for all samples sent on plates.
    Applicable to all order types.
    """
    errors: list[ContainerNameMissingError] = []
    for sample_index, sample in order.enumerated_samples:
        if is_container_name_missing(sample=sample):
            error = ContainerNameMissingError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_index_number_in_range(
    order: OrderWithIndexedSamples, **kwargs
) -> list[IndexNumberOutOfRangeError]:
    errors: list[IndexNumberOutOfRangeError] = []
    for sample_index, sample in order.enumerated_samples:
        if is_index_number_out_of_range(sample):
            error = IndexNumberOutOfRangeError(sample_index=sample_index, index=sample.index)
            errors.append(error)
    return errors


def validate_index_number_required(
    order: OrderWithIndexedSamples, **kwargs
) -> list[IndexNumberMissingError]:
    errors: list[IndexNumberMissingError] = []
    for sample_index, sample in order.enumerated_samples:
        if is_index_number_missing(sample):
            error = IndexNumberMissingError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_non_control_sample_names_available(
    order: OrderWithControlSamples, store: Store, **kwargs
) -> list[SampleNameNotAvailableControlError]:
    """
    Validate that non-control sample names do not exists in the database under the same customer.
    Applicable to all orders with control samples.
    """
    errors: list[SampleNameNotAvailableControlError] = get_sample_name_not_available_errors(
        order=order, store=store, has_order_control=True
    )
    return errors


def validate_pools_contain_one_application(
    order: OrderWithIndexedSamples, **kwargs
) -> list[PoolApplicationError]:
    """
    Validate that the pools in the order contain only samples with the same application.
    Only applicable to order types with indexed samples (RML and Fluffy).
    """
    errors: list[PoolApplicationError] = []
    for pool, enumerated_samples in order.enumerated_pools.items():
        samples: list[IndexedSample] = [sample for _, sample in enumerated_samples]
        if has_multiple_applications(samples):
            for sample_index, _ in enumerated_samples:
                error = PoolApplicationError(sample_index=sample_index, pool_name=pool)
                errors.append(error)
    return errors


def validate_pools_contain_one_priority(
    order: OrderWithIndexedSamples, **kwargs
) -> list[PoolPriorityError]:
    """
    Validate that the pools in the order contain only samples with the same priority.
    Only applicable to order types with indexed samples (RML and Fluffy).
    """
    errors: list[PoolPriorityError] = []
    for pool, enumerated_samples in order.enumerated_pools.items():
        samples: list[IndexedSample] = [sample for _, sample in enumerated_samples]
        if has_multiple_priorities(samples):
            for sample_index, _ in enumerated_samples:
                error = PoolPriorityError(sample_index=sample_index, pool_name=pool)
                errors.append(error)
    return errors


def validate_sample_names_available(
    order: OrderWithSamples, store: Store, **kwargs
) -> list[SampleError]:
    """
    Validate that the sample names do not exists in the database under the same customer.
    Applicable to all orders without control samples.
    """
    errors: list[SampleError] = get_sample_name_not_available_errors(
        order=order, store=store, has_order_control=False
    )
    return errors


def validate_sample_names_unique(
    order: OrderWithSamples, **kwargs
) -> list[SampleNameRepeatedError]:
    """
    Validate that all the sample names are unique within the order.
    Applicable to all order types.
    """
    sample_indices: list[int] = get_indices_for_repeated_sample_names(order)
    return [SampleNameRepeatedError(sample_index=sample_index) for sample_index in sample_indices]


def validate_tube_container_name_unique(
    order: OrderWithSamples,
    **kwargs,
) -> list[ContainerNameRepeatedError]:
    """
    Validate that the container names are unique for tube samples within the order.
    Applicable to all order types.
    """
    errors: list[ContainerNameRepeatedError] = []
    repeated_container_name_indices: list = get_indices_for_tube_repeated_container_name(order)
    for sample_index in repeated_container_name_indices:
        error = ContainerNameRepeatedError(sample_index=sample_index)
        errors.append(error)
    return errors


def validate_volume_interval(order: OrderWithSamples, **kwargs) -> list[InvalidVolumeError]:
    """
    Validate that the volume of all samples is within the allowed interval.
    Applicable to all order types.
    """
    errors: list[InvalidVolumeError] = []
    for sample_index, sample in order.enumerated_samples:
        if is_volume_invalid(sample):
            error = InvalidVolumeError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_volume_required(order: OrderWithSamples, **kwargs) -> list[VolumeRequiredError]:
    """
    Validate that all samples have a volume if they are in a container.
    Applicable to all order types.
    """
    errors: list[VolumeRequiredError] = []
    for sample_index, sample in order.enumerated_samples:
        if is_volume_missing(sample):
            error = VolumeRequiredError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_wells_contain_at_most_one_sample(
    order: OrderWithSamples,
    **kwargs,
) -> list[OccupiedWellError]:
    """
    Validate that the wells in the order contain at most one sample.
    Applicable to all order types with non-indexed samples.
    """
    plate_samples = PlateSamplesValidator(order)
    return plate_samples.get_occupied_well_errors()


def validate_well_position_format(order: OrderWithSamples, **kwargs) -> list[WellFormatError]:
    """
    Validate that the well positions of all samples sent in plates have the correct format.
    Applicable to all order types with non-indexed samples.
    """
    errors: list[WellFormatError] = []
    for sample_index, sample in order.enumerated_samples:
        if is_invalid_well_format(sample=sample):
            error = WellFormatError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_well_position_rml_format(
    order: OrderWithIndexedSamples, **kwargs
) -> list[WellFormatRmlError]:
    """
    Validate that the well positions of all indexed samples have the correct format.
    Applicable to all order types with indexed samples.
    """
    errors: list[WellFormatRmlError] = []
    for sample_index, sample in order.enumerated_samples:
        if sample.well_position_rml and is_invalid_well_format_rml(sample=sample):
            error = WellFormatRmlError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_well_positions_required(
    order: OrderWithSamples,
    **kwargs,
) -> list[WellPositionMissingError]:
    """
    Validate that all samples sent in plates have well positions.
    Applicable to all order types with non-indexed samples
    """
    plate_samples = PlateSamplesValidator(order)
    return plate_samples.get_well_position_missing_errors()


def validate_well_positions_required_rml(
    order: OrderWithIndexedSamples, **kwargs
) -> list[WellPositionRmlMissingError]:
    """
    Validate that all indexed samples have well positions.
    Applicable to all order types with indexed samples.
    """
    errors: list[WellPositionRmlMissingError] = []
    for sample_index, sample in order.enumerated_samples:
        if sample.is_on_plate and not sample.well_position_rml:
            error = WellPositionRmlMissingError(sample_index=sample_index)
            errors.append(error)
    return errors


def validate_source_comment_required(
    order: FastqOrder | MetagenomeOrder | TaxprofilerOrder | PacbioOrder, **kwargs
) -> list[MissingSourceCommentError]:
    errors: list[MissingSourceCommentError] = []
    for sample_index, sample in order.enumerated_samples:
        if sample.source == "other" and not sample.source_comment:
            error = MissingSourceCommentError(sample_index=sample_index)
            errors.append(error)
    return errors



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.rml.constants import RMLDeliveryType
from cg.services.orders.validation.order_types.rml.models.sample import RMLSample


class RMLOrder(OrderWithSamples[RMLSample]):
    delivery_type: RMLDeliveryType

    @property
    def pools(self) -> dict[str, list[RMLSample]]:
        """Return a dictionary matching pool names and their respective samples."""
        pools: dict[str, list[RMLSample]] = {}
        for sample in self.samples:
            if sample.pool not in pools:
                pools[sample.pool] = [sample]
            else:
                pools[sample.pool].append(sample)
        return pools

    @property
    def enumerated_pools(self) -> dict[str, list[tuple[int, RMLSample]]]:
        """Return the pool dictionary with indexes for the samples to map them to validation errors."""
        pools: dict[str, list[tuple[int, RMLSample]]] = {}
        for sample_index, sample in self.enumerated_samples:
            if sample.pool not in pools:
                pools[sample.pool] = [(sample_index, sample)]
            else:
                pools[sample.pool].append((sample_index, sample))
        return pools



from flask import Blueprint, jsonify

from cg.server.endpoints.error_handler import handle_missing_entries
from cg.server.endpoints.sequencing_run.dtos import PacbioSmrtCellMetricsResponse
from cg.server.endpoints.utils import before_request
from cg.server.ext import pacbio_sequencing_runs_service

PACBIO_SMRT_CELL_METRICS_BLUEPRINT = Blueprint(
    "pacbio_smrt_cell_metrics", __name__, url_prefix="/api/v1/pacbio_smrt_cell_metrics"
)
PACBIO_SMRT_CELL_METRICS_BLUEPRINT.before_request(before_request)


@PACBIO_SMRT_CELL_METRICS_BLUEPRINT.route("/<run_id>", methods=["GET"])
@handle_missing_entries
def get_smrt_cell_metrics(run_id: str):
    response: PacbioSmrtCellMetricsResponse = (
        pacbio_sequencing_runs_service.get_sequencing_runs_by_run_id(run_id)
    )
    return jsonify(response.model_dump())



import logging
from pathlib import Path

from cg.services.run_devices.abstract_classes import PostProcessingService
from cg.services.run_devices.constants import POST_PROCESSING_COMPLETED
from cg.services.run_devices.error_handler import handle_post_processing_errors
from cg.services.run_devices.exc import (
    PostProcessingError,
    PostProcessingRunDataGeneratorError,
    PostProcessingRunFileManagerError,
    PostProcessingStoreDataError,
    PostProcessingStoreFileError,
)
from cg.services.run_devices.pacbio.data_storage_service.pacbio_store_service import (
    PacBioStoreService,
)
from cg.services.run_devices.pacbio.housekeeper_service.pacbio_houskeeper_service import (
    PacBioHousekeeperService,
)
from cg.services.run_devices.pacbio.run_data_generator.pacbio_run_data_generator import (
    PacBioRunDataGenerator,
)
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.services.run_devices.pacbio.run_validator.pacbio_run_validator import PacBioRunValidator

LOG = logging.getLogger(__name__)


class PacBioPostProcessingService(PostProcessingService):
    """Service for handling post-processing of PacBio sequencing runs."""

    def __init__(
        self,
        run_validator: PacBioRunValidator,
        run_data_generator: PacBioRunDataGenerator,
        hk_service: PacBioHousekeeperService,
        store_service: PacBioStoreService,
        sequencing_dir: str,
    ):
        self.run_validator: PacBioRunValidator = run_validator
        self.run_data_generator: PacBioRunDataGenerator = run_data_generator
        self.hk_service: PacBioHousekeeperService = hk_service
        self.store_service: PacBioStoreService = store_service
        self.sequencing_dir: str = sequencing_dir

    @handle_post_processing_errors(
        to_except=(
            PostProcessingStoreDataError,
            PostProcessingRunDataGeneratorError,
            PostProcessingStoreFileError,
        ),
        to_raise=PostProcessingError,
    )
    def post_process(self, run_full_name: str, dry_run: bool = False) -> None:
        LOG.info(f"Starting Pacbio post-processing for run: {run_full_name}")
        run_data: PacBioRunData = self.run_data_generator.get_run_data(
            run_full_name=run_full_name, sequencing_dir=self.sequencing_dir
        )
        self.run_validator.ensure_post_processing_can_start(run_data)
        self.store_service.store_post_processing_data(run_data=run_data, dry_run=dry_run)
        self.hk_service.store_files_in_housekeeper(run_data=run_data, dry_run=dry_run)
        self._touch_post_processing_complete(run_data=run_data, dry_run=dry_run)

    def is_run_processed(self, run_full_name: str) -> bool:
        """Check if a run has been post-processed."""
        processing_complete_file = Path(
            self.sequencing_dir, run_full_name, POST_PROCESSING_COMPLETED
        )
        return processing_complete_file.exists()

    def can_post_processing_start(self, run_full_name: str) -> bool:
        """Makes sure that all SMRT cells in the sequencing run are ready to be post-processed."""
        LOG.debug(f"Checking if Pacbio post-processing can start for run: {run_full_name}")
        parent_directory: Path = Path(self.sequencing_dir, run_full_name).parent
        all_smrt_cells_are_ready: bool = all(
            self.is_smrt_cell_ready_for_post_processing(f"{parent_directory.name}/{smrt_cell.name}")
            for smrt_cell in parent_directory.iterdir()
        )
        return all_smrt_cells_are_ready

    def is_smrt_cell_ready_for_post_processing(self, smrt_cell_full_name: str) -> bool:
        LOG.debug(f"Checking if Pacbio SMRT-cell {smrt_cell_full_name} is ready for postprocessing")
        try:
            run_data: PacBioRunData = self.run_data_generator.get_run_data(
                run_full_name=smrt_cell_full_name, sequencing_dir=self.sequencing_dir
            )
            self.run_validator.validate_run_files(run_data)
        except PostProcessingRunFileManagerError as error:
            LOG.debug(
                f"Run {smrt_cell_full_name} is not ready for post-processing. {error.args[0]}."
            )
            return False
        return True



from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.mip_rna.constants import MIPRNADeliveryType
from cg.services.orders.validation.order_types.mip_rna.models.case import MIPRNACase
from cg.services.orders.validation.order_types.mip_rna.models.sample import MIPRNASample


class MIPRNAOrder(OrderWithCases[MIPRNACase, MIPRNASample]):
    delivery_type: MIPRNADeliveryType



from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.tomte.constants import TomteDeliveryType
from cg.services.orders.validation.order_types.tomte.models.case import TomteCase
from cg.services.orders.validation.order_types.tomte.models.sample import TomteSample


class TomteOrder(OrderWithCases[TomteCase, TomteSample]):
    delivery_type: TomteDeliveryType



"""Constants related to PacBio sequencing."""

from cg.constants import FileExtensions


class PacBioDirsAndFiles:
    BARCODES_REPORT: str = "barcodes.report.json"
    CCS_REPORT_SUFFIX: str = "ccs_report.json"
    CONTROL_REPORT: str = "control.report.json"
    LOADING_REPORT: str = "loading.report.json"
    HIFI_READS: str = "hifi_reads"
    RAW_DATA_REPORT: str = "raw_data.report.json"
    SMRTLINK_DATASETS_REPORT: str = "smrtlink-datasets.json"
    STATISTICS_DIR: str = "statistics"
    UNZIPPED_REPORTS_DIR: str = "unzipped_reports"
    METADATA_DIR: str = "metadata"
    METADATA_FILE_SUFFIX: str = "metadata.xml"
    RUN_IS_VALID: str = "is_valid"


class MetricsFileFields:
    ATTRIBUTES: str = "attributes"
    COLUMNS: str = "columns"
    ID: str = "id"
    TABLES: str = "tables"
    VALUE: str = "value"
    VALUES: str = "values"


class CCSAttributeAliases:
    HIFI_READS: str = "ccs_processing.number_of_ccs_reads_q20"
    HIFI_YIELD: str = "ccs_processing.total_number_of_ccs_bases_q20"
    HIFI_MEAN_READ_LENGTH: str = "ccs_processing.mean_ccs_readlength_q20"
    HIFI_MEDIAN_READ_LENGTH: str = "ccs_processing.median_ccs_readlength_q20"
    HIFI_READ_LENGTH_N50: str = "ccs_processing.ccs_readlength_n50_q20"
    HIFI_MEDIAN_READ_QUALITY: str = "ccs_processing.median_qv_q20"
    PERCENT_Q30: str = "ccs_processing.base_percentage_q30"
    FAILED_READS: str = "ccs_processing.number_of_ccs_reads_lq"
    FAILED_YIELD: str = "ccs_processing.total_number_of_ccs_bases_lq"
    FAILED_MEAN_READ_LENGTH: str = "ccs_processing.mean_ccs_readlength_lq"


class ControlAttributeAliases:
    NUMBER_OF_READS: str = "control.reads_n"
    MEAN_READ_LENGTH: str = "control.readlength_mean"
    PERCENT_MEAN_READ_CONCORDANCE: str = "control.concordance_mean"
    PERCENT_MODE_READ_CONCORDANCE: str = "control.concordance_mode"


class LoadingAttributesAliases:
    PRODUCTIVE_ZMWS: str = "loading_xml_report.productive_zmws"
    P_0: str = "loading_xml_report.productivity_0_n"
    P_1: str = "loading_xml_report.productivity_1_n"
    P_2: str = "loading_xml_report.productivity_2_n"


class PolymeraseDataAttributeAliases:
    MEAN_READ_LENGTH: str = "raw_data_report.read_length"
    READ_LENGTH_N50: str = "raw_data_report.read_n50"
    MEAN_LONGEST_SUBREAD_LENGTH: str = "raw_data_report.insert_length"
    LONGEST_SUBREAD_LENGTH_N50: str = "raw_data_report.insert_n50"


class SmrtLinkDatabasesAliases:
    BIO_SAMPLE_NAME: str = "bioSampleName"
    CELL_ID: str = "cellId"
    CELL_INDEX: str = "cellIndex"
    INSTRUMENT_NAME: str = "instrumentName"
    MOVIE_NAME: str = "metadataContextId"
    PATH: str = "path"
    RUN_COMPLETED_AT = "createdAt"
    WELL_NAME: str = "wellName"
    WELL_SAMPLE_NAME: str = "wellSampleName"


class BarcodeMetricsAliases:
    BARCODED_HIFI_READS: str = "barcode.n_barcoded_reads"
    BARCODED_HIFI_READS_PERCENTAGE: str = "barcode.percent_barcoded_reads"
    BARCODED_HIFI_YIELD: str = "barcode.barcoded_bases"
    BARCODED_HIFI_YIELD_PERCENTAGE: str = "barcode.percent_barcoded_bases"
    BARCODED_MEAN_READ_LENGTH: str = "barcode.mean_read_length"
    UNBARCODED_HIFI_READS: str = "barcode.n_unbarcoded_reads"
    UNBARCODED_HIFI_YIELD: str = "barcode.unbarcoded_bases"
    UNBARCODED_HIFI_MEAN_READ_LENGTH: str = "barcode.mean_unbarcoded_read_length"


class SampleMetricsAliases:
    BARCODE_NAME: str = "barcode.barcode_table.barcode"
    HIFI_MEAN_READ_LENGTH: str = "barcode.barcode_table.mean_read_length"
    HIFI_READ_QUALITY: str = "barcode.barcode_table.median_read_quality"
    HIFI_READS: str = "barcode.barcode_table.number_of_reads"
    HIFI_YIELD: str = "barcode.barcode_table.number_of_bases"
    POLYMERASE_READ_LENGTH: str = "barcode.barcode_table.mean_polymerase_read_length"
    SAMPLE_INTERNAL_ID: str = "barcode.barcode_table.biosample"


class PacBioHousekeeperTags:
    BARCODES_REPORT: str = "barcodes-report"
    CCS_REPORT: str = "ccs-report"
    CONTROL_REPORT: str = "control-report"
    LOADING_REPORT: str = "loading-report"
    RAWDATA_REPORT: str = "raw-data-report"
    DATASETS_REPORT: str = "datasets-report"
    METADATA: str = "smrt-link-metadata"


class PacBioBundleTypes:
    SAMPLE: str = "sample"
    SMRT_CELL: str = "smrt_cell"


file_pattern_to_bundle_type: dict[str, str] = {
    PacBioDirsAndFiles.BARCODES_REPORT: PacBioBundleTypes.SMRT_CELL,
    PacBioDirsAndFiles.CONTROL_REPORT: PacBioBundleTypes.SMRT_CELL,
    f".*{PacBioDirsAndFiles.CCS_REPORT_SUFFIX}$": PacBioBundleTypes.SMRT_CELL,
    PacBioDirsAndFiles.LOADING_REPORT: PacBioBundleTypes.SMRT_CELL,
    PacBioDirsAndFiles.RAW_DATA_REPORT: PacBioBundleTypes.SMRT_CELL,
    PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT: PacBioBundleTypes.SMRT_CELL,
    f"{PacBioDirsAndFiles.HIFI_READS}.*{FileExtensions.BAM}$": PacBioBundleTypes.SAMPLE,
    f".*{PacBioDirsAndFiles.METADATA_FILE_SUFFIX}$": PacBioBundleTypes.SMRT_CELL,
}

ZIPPED_REPORTS_PATTERN: str = "*reports.zip"
MANIFEST_FILE_PATTERN: str = "*.transferdone"



"""Helper functions for the back-up of Illumina flow cells."""

from datetime import datetime
from operator import attrgetter
from pathlib import Path

from cg.constants import FileExtensions
from cg.meta.backup.backup import LOG
from cg.services.illumina.backup.models import PdcEncryptionKey, PdcSequencingFile
from cg.utils.date import convert_string_to_datetime_object


class DsmcOutput:
    DATE_COLUMN_INDEX = 2
    TIME_COLUMN_INDEX = 3
    PATH_COLUMN_INDEX = 4


def get_latest_archived_sequencing_run_path(dsmc_output: list[str]) -> Path | None:
    """Get the path of the archived sequencing run from a PDC query."""
    validated_sequencing_paths: list[PdcSequencingFile] = _parse_dsmc_output_sequencing_path(
        dsmc_output
    )
    archived_run: PdcSequencingFile | None = _get_latest_sequencing_run_in_list(
        validated_sequencing_paths
    )
    if archived_run:
        LOG.info(f"Sequencing run found: {archived_run}")
        return archived_run.path


def get_latest_archived_encryption_key_path(dsmc_output: list[str]) -> Path | None:
    """Get the encryption key for the archived sequencing run from a PDC query."""
    validated_encryption_keys: list[PdcEncryptionKey] = _parse_dsmc_output_key_path(dsmc_output)

    archived_encryption_key: PdcEncryptionKey | None = _get_latest_pdc_encryption_key(
        validated_encryption_keys
    )
    if archived_encryption_key:
        LOG.info(f"Encryption key found: {archived_encryption_key}")
        return archived_encryption_key.path


def _parse_dsmc_output_sequencing_path(dsmc_output: list[str]) -> list[PdcSequencingFile]:
    """Parses the DSMC command output to extract validated sequencing paths."""
    validated_responses: list[PdcSequencingFile] = []
    for line in dsmc_output:
        if is_pdc_sequencing_path(line):
            parts: list[str] = line.split()

            file_date_time: datetime = convert_string_to_datetime_object(
                f"{parts[DsmcOutput.DATE_COLUMN_INDEX]} {parts[DsmcOutput.TIME_COLUMN_INDEX]}"
            )

            query_response = PdcSequencingFile(
                date_time=file_date_time,
                path=Path(parts[DsmcOutput.PATH_COLUMN_INDEX]),
            )
            validated_responses.append(query_response)

    return validated_responses


def _parse_dsmc_output_key_path(dsmc_output: list[str]) -> list[PdcEncryptionKey]:
    """Parses the DSMC command output to extract validated encryption keys."""
    validated_responses: list[PdcEncryptionKey] = []
    for line in dsmc_output:
        if is_pdc_encryption_key(line):
            parts: list[str] = line.split()

            file_date_time: datetime = convert_string_to_datetime_object(
                f"{parts[DsmcOutput.DATE_COLUMN_INDEX]} {parts[DsmcOutput.TIME_COLUMN_INDEX]}"
            )

            query_response = PdcEncryptionKey(
                date_time=file_date_time,
                path=Path(parts[DsmcOutput.PATH_COLUMN_INDEX]),
            )

            validated_responses.append(query_response)

    return validated_responses


def _get_latest_sequencing_run_in_list(
    pdc_files: list[PdcSequencingFile],
) -> PdcSequencingFile | None:
    """Return the most recent sequencing run based on the 'dateTime' attribute, or None if no files are present."""

    if not pdc_files:
        return None

    latest_file: PdcSequencingFile | None = max(pdc_files, key=attrgetter("date_time"))

    return latest_file


def _get_latest_pdc_encryption_key(pdc_files: list[PdcEncryptionKey]) -> PdcEncryptionKey | None:
    """Return the latest encryption key or None if no keys are present."""

    if not pdc_files:
        return None

    latest_file: PdcEncryptionKey | None = max(pdc_files, key=attrgetter("date_time"))

    return latest_file


def is_pdc_sequencing_path(line: str) -> bool:
    return FileExtensions.TAR in line and FileExtensions.GZIP in line and FileExtensions.GPG in line


def is_pdc_encryption_key(line: str) -> bool:
    return (
        FileExtensions.KEY in line
        and FileExtensions.GPG in line
        and FileExtensions.GZIP not in line
    )



from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from cg.store.models import BedVersion


def filter_bed_version_by_file_name(
    bed_versions: Query, bed_version_file_name: str, **kwargs
) -> Query:
    """Return beds by file name."""
    return bed_versions.filter(BedVersion.filename == bed_version_file_name)


class BedVersionFilter(Enum):
    """Define BED version filter functions."""

    BY_FILE_NAME: Callable = filter_bed_version_by_file_name


def apply_bed_version_filter(
    bed_versions: Query,
    filter_functions: list[Callable],
    bed_version_file_name: str | None = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filter_functions:
        bed_versions: Query = function(
            bed_versions=bed_versions,
            bed_version_file_name=bed_version_file_name,
        )
    return bed_versions



from enum import StrEnum

from cg.constants import DataDelivery


class TaxprofilerDeliveryType(StrEnum):
    ANALYSIS = DataDelivery.ANALYSIS_FILES
    FASTQ_ANALYSIS = DataDelivery.FASTQ_ANALYSIS
    NO_DELIVERY = DataDelivery.NO_DELIVERY



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.taxprofiler.constants import TaxprofilerDeliveryType
from cg.services.orders.validation.order_types.taxprofiler.models.sample import TaxprofilerSample


class TaxprofilerOrder(OrderWithSamples[TaxprofilerSample]):
    delivery_type: TaxprofilerDeliveryType



from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.balsamic.constants import BalsamicDeliveryType
from cg.services.orders.validation.order_types.balsamic.models.case import BalsamicCase
from cg.services.orders.validation.order_types.balsamic.models.sample import BalsamicSample


class BalsamicOrder(OrderWithCases[BalsamicCase, BalsamicSample]):
    delivery_type: BalsamicDeliveryType



import logging
from functools import wraps
from http import HTTPStatus

from flask import jsonify

from cg.store.exc import EntryNotFoundError

LOG = logging.getLogger(__name__)


def handle_missing_entries(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except EntryNotFoundError as error:
            LOG.error(error)
            return jsonify(error=str(error)), HTTPStatus.NOT_FOUND
        except Exception as error:
            LOG.error(f"Unexpected error in endpoint: {error}")
            return (
                jsonify(error="An error occurred while processing your request."),
                HTTPStatus.INTERNAL_SERVER_ERROR,
            )

    return wrapper



"""CLI support to create config and/or start BALSAMIC"""

import logging
from typing import cast

import rich_click as click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.cli.workflow.balsamic.base import report_deliver, store, store_available, store_housekeeper
from cg.cli.workflow.balsamic.options import OPTION_PANEL_BED, OPTION_WORKFLOW_PROFILE
from cg.cli.workflow.commands import ARGUMENT_CASE_ID, link, resolve_compression
from cg.constants.constants import Workflow
from cg.meta.workflow.balsamic_umi import BalsamicUmiAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.services.analysis_starter.analysis_starter import AnalysisStarter
from cg.services.analysis_starter.configurator.implementations.balsamic import BalsamicConfigurator
from cg.services.analysis_starter.factories.configurator_factory import ConfiguratorFactory
from cg.services.analysis_starter.factories.starter_factory import AnalysisStarterFactory

LOG = logging.getLogger(__name__)


@click.group(
    "balsamic-umi",
    invoke_without_command=True,
    context_settings=CLICK_CONTEXT_SETTINGS,
)
@click.pass_context
def balsamic_umi(context: click.Context):
    """Cancer analysis workflow"""
    if context.invoked_subcommand is None:
        click.echo(context.get_help())
        return None
    config = context.obj
    context.obj.meta_apis["analysis_api"] = BalsamicUmiAnalysisAPI(config=config)


balsamic_umi.add_command(resolve_compression)
balsamic_umi.add_command(link)
balsamic_umi.add_command(report_deliver)
balsamic_umi.add_command(store_housekeeper)
balsamic_umi.add_command(store)
balsamic_umi.add_command(store_available)


@balsamic_umi.command("config-case")
@OPTION_PANEL_BED
@ARGUMENT_CASE_ID
@click.pass_obj
def config_case(cg_config: CGConfig, case_id: str, panel_bed: str | None):
    """Configure a Balsamic-UMI case so that it is ready to be run.

    \b
    Creates the case config file:
        - CASE_ID.json
    """
    factory = ConfiguratorFactory(cg_config)
    configurator = cast(BalsamicConfigurator, factory.get_configurator(Workflow.BALSAMIC_UMI))
    configurator.configure(case_id=case_id, panel_bed=panel_bed)


@balsamic_umi.command("run")
@OPTION_WORKFLOW_PROFILE
@ARGUMENT_CASE_ID
@click.pass_obj
def run(cg_config: CGConfig, case_id: str, workflow_profile: click.Path | None):
    """
    Run a preconfigured Balsamic-UMI case.

    \b
    Assumes that case config file exist in the case run directory:
        - CASE_ID.json
    """
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(
        Workflow.BALSAMIC_UMI
    )
    analysis_starter.run(case_id=case_id, workflow_profile=workflow_profile)


@balsamic_umi.command("start")
@OPTION_PANEL_BED
@OPTION_WORKFLOW_PROFILE
@ARGUMENT_CASE_ID
@click.pass_obj
def start(
    cg_config: CGConfig,
    case_id: str,
    panel_bed: str | None,
    workflow_profile: click.Path | None,
):
    """
    Starts a Balsamic-UMI case.

    \b
    Configures the case and creates the case config file:
        - CASE_ID.json
    and submits the job to slurm.
    """
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(
        Workflow.BALSAMIC_UMI
    )
    analysis_starter.start(case_id=case_id, workflow_profile=workflow_profile, panel_bed=panel_bed)


@balsamic_umi.command("start-available")
@click.pass_obj
def start_available(cg_config: CGConfig):
    """
    Starts all available Balsamic-UMI cases.

    \b
    Configures the individual case and creates its case config file:
        - CASE_ID.json
    and submits the job to slurm.
    """
    LOG.info("Starting Balsamic-UMI workflow for all available cases.")
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter = factory.get_analysis_starter_for_workflow(Workflow.BALSAMIC_UMI)
    succeeded: bool = analysis_starter.start_available()
    if not succeeded:
        raise click.Abort



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.microsalt.constants import MicrosaltDeliveryType
from cg.services.orders.validation.order_types.microsalt.models.sample import MicrosaltSample


class MicrosaltOrder(OrderWithSamples[MicrosaltSample]):
    delivery_type: MicrosaltDeliveryType



import logging
from datetime import datetime

from cg.models.orders.sample_base import PriorityEnum, SexEnum, StatusEnum
from cg.services.orders.constants import ORDER_TYPE_WORKFLOW_MAP
from cg.services.orders.lims_service.service import OrderLimsService
from cg.services.orders.storing.service import StoreOrderService
from cg.services.orders.validation.models.order_aliases import OrderWithIndexedSamples
from cg.services.orders.validation.models.sample_aliases import IndexedSample
from cg.store.models import ApplicationVersion, Case, CaseSample, Customer, Order, Pool, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class StorePoolOrderService(StoreOrderService):
    """
    Service for storing generic orders in StatusDB and Lims.
    This class is used to store orders for the following order types:
    - Fluffy / NIPT samples
    - RML samples
    """

    def __init__(self, status_db: Store, lims_service: OrderLimsService):
        self.status_db = status_db
        self.lims = lims_service

    def store_order(self, order: OrderWithIndexedSamples) -> dict:
        project_data, lims_map = self.lims.process_lims(
            samples=order.samples,
            customer=order.customer,
            ticket=order._generated_ticket_id,
            order_name=order.name,
            workflow=ORDER_TYPE_WORKFLOW_MAP[order.order_type],
            delivery_type=order.delivery_type,
            skip_reception_control=order.skip_reception_control,
        )
        self._fill_in_sample_ids(samples=order.samples, lims_map=lims_map)
        new_records: list[Pool] = self.store_order_data_in_status_db(order=order)
        return {"project": project_data, "records": new_records}

    def store_order_data_in_status_db(self, order: OrderWithIndexedSamples) -> list[Pool]:
        """Store pools in the status database."""
        db_order: Order = self._create_db_order(order=order)
        new_pools: list[Pool] = []
        with self.status_db.no_autoflush_context():
            for pool in order.pools.items():
                db_case: Case = self._create_db_case_for_pool(
                    order=order,
                    pool=pool,
                    customer=db_order.customer,
                    ticket_id=str(db_order.ticket_id),
                )
                db_pool: Pool = self._create_db_pool(
                    pool=pool,
                    order_name=order.name,
                    ticket_id=str(db_order.ticket_id),
                    customer=db_order.customer,
                )
                for sample in pool[1]:
                    db_sample: Sample = self._create_db_sample(
                        sample=sample,
                        order_name=order.name,
                        ticket_id=str(db_order.ticket_id),
                        customer=db_order.customer,
                        application_version=db_pool.application_version,
                    )
                    case_sample: CaseSample = self.status_db.relate_sample(
                        case=db_case,
                        sample=db_sample,
                        status=StatusEnum.unknown,
                        should_deliver_sample=True,
                    )
                    self.status_db.add_multiple_items_to_store([db_sample, case_sample])
                new_pools.append(db_pool)
                db_order.cases.append(db_case)
                self.status_db.add_multiple_items_to_store([db_pool, db_case])
        self.status_db.add_item_to_store(db_order)
        self.status_db.commit_to_store()
        return new_pools

    @staticmethod
    def create_case_name(ticket: str, pool_name: str) -> str:
        return f"{ticket}-{pool_name}"

    def _get_application_version_from_pool_samples(
        self, pool_samples: list[IndexedSample]
    ) -> ApplicationVersion:
        """
        Return the application version for a pool by taking the app tag of the first sample of
        the pool. The validation guarantees that all samples in a pool have the same application.
        """
        app_tag: str = pool_samples[0].application
        application_version: ApplicationVersion = (
            self.status_db.get_current_application_version_by_tag(tag=app_tag)
        )
        return application_version

    @staticmethod
    def _get_priority_from_pool_samples(pool_samples: list[IndexedSample]) -> PriorityEnum:
        """
        Return the priority of the pool by taking the priority of the first sample of the pool.
        The validation guarantees that all samples in a pool have the same priority.
        """
        return pool_samples[0].priority

    def _create_db_order(self, order: OrderWithIndexedSamples) -> Order:
        """Return an Order database object."""
        ticket_id: int = order._generated_ticket_id
        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=order.customer
        )
        return self.status_db.add_order(customer=customer, ticket_id=ticket_id)

    def _create_db_case_for_pool(
        self,
        order: OrderWithIndexedSamples,
        pool: tuple[str, list[IndexedSample]],
        customer: Customer,
        ticket_id: str,
    ) -> Case:
        """Return a Case database object for a pool."""
        case_name: str = self.create_case_name(ticket=ticket_id, pool_name=pool[0])
        case = self.status_db.add_case(
            data_analysis=ORDER_TYPE_WORKFLOW_MAP[order.order_type],
            data_delivery=order.delivery_type,
            name=case_name,
            priority=self._get_priority_from_pool_samples(pool_samples=pool[1]),
            ticket=ticket_id,
        )
        case.customer = customer
        return case

    def _create_db_pool(
        self,
        pool: tuple[str, list[IndexedSample]],
        order_name: str,
        ticket_id: str,
        customer: Customer,
    ) -> Pool:
        """Return a Pool database object."""
        application_version: ApplicationVersion = self._get_application_version_from_pool_samples(
            pool_samples=pool[1]
        )
        return self.status_db.add_pool(
            application_version=application_version,
            customer=customer,
            name=pool[0],
            order=order_name,
            ordered=datetime.now(),
            ticket=ticket_id,
        )

    def _create_db_sample(
        self,
        sample: IndexedSample,
        order_name: str,
        ticket_id: str,
        customer: Customer,
        application_version: ApplicationVersion,
    ) -> Sample:
        """Return a Sample database object."""
        return self.status_db.add_sample(
            name=sample.name,
            customer=customer,
            application_version=application_version,
            sex=SexEnum.unknown,
            comment=sample.comment,
            control=sample.control,
            internal_id=sample._generated_lims_id,
            order=order_name,
            ordered=datetime.now(),
            original_ticket=ticket_id,
            priority=sample.priority,
            no_invoice=True,
        )



import logging
from typing import Callable

from cg.constants.constants import SequencingQCStatus
from cg.services.sequencing_qc_service.quality_checks.checks import (
    get_sample_sequencing_quality_check,
    get_sequencing_quality_check_for_case,
    run_quality_checks,
)
from cg.services.sequencing_qc_service.utils import qc_bool_to_status
from cg.store.models import Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class SequencingQCService:
    def __init__(self, store: Store):
        self.store = store

    def run_sequencing_qc(self) -> bool:
        """
        Run QC for samples in pending or failed cases and store the aggregated score on each case.
        Return True if all checks could run succesfully
        Return False if at least one of the checks raised an exception
        """
        cases: list[Case] = self.store.get_cases_for_sequencing_qc()
        all_checks_ran_succesfully: bool = True

        for case in cases:
            LOG.debug(f"Performing sequencing QC for case: {case.internal_id}")
            try:
                passes_qc: bool = self.case_pass_sequencing_qc(case)
                qc_status: SequencingQCStatus = qc_bool_to_status(passes_qc)
                self.store.update_sequencing_qc_status(case=case, status=qc_status)
                LOG.info(f"Sequencing QC status for case {case.internal_id}: {qc_status}")
            except Exception as e:
                LOG.error(f"Error found during sequencing QC of case: {case.internal_id}: {e}")
                all_checks_ran_succesfully = False

        return all_checks_ran_succesfully

    @staticmethod
    def case_pass_sequencing_qc(case: Case) -> bool:
        """Run the sequencing QC for the case or sample."""
        sequencing_quality_check: Callable = get_sequencing_quality_check_for_case(case)
        return run_quality_checks(quality_checks=[sequencing_quality_check], case=case)

    @staticmethod
    def sample_pass_sequencing_qc(sample: Sample) -> bool:
        """Run the sequencing QC for a sample."""
        sample_sequencing_quality_check: Callable = get_sample_sequencing_quality_check()
        return run_quality_checks(quality_checks=[sample_sequencing_quality_check], sample=sample)



from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.mip_dna.constants import MIPDNADeliveryType
from cg.services.orders.validation.order_types.mip_dna.models.case import MIPDNACase
from cg.services.orders.validation.order_types.mip_dna.models.sample import MIPDNASample


class MIPDNAOrder(OrderWithCases[MIPDNACase, MIPDNASample]):
    delivery_type: MIPDNADeliveryType



from typing import Generic, TypeVar

from pydantic import BaseModel, Discriminator, Field, Tag, model_validator
from typing_extensions import Annotated

from cg.constants.priority import PriorityTerms
from cg.models.orders.sample_base import NAME_PATTERN
from cg.services.orders.validation.models.discriminators import has_internal_id
from cg.services.orders.validation.models.existing_sample import ExistingSample
from cg.services.orders.validation.models.sample import Sample
from cg.store.models import Sample as DbSample
from cg.store.store import Store

SampleType = TypeVar("SampleType", bound=Sample)


class Case(BaseModel, Generic[SampleType]):
    name: str = Field(pattern=NAME_PATTERN, min_length=2, max_length=128)
    priority: PriorityTerms = PriorityTerms.STANDARD
    samples: list[
        Annotated[
            Annotated[SampleType, Tag("new")] | Annotated[ExistingSample, Tag("existing")],
            Discriminator(has_internal_id),
        ]
    ]

    @property
    def is_new(self) -> bool:
        return True

    @property
    def enumerated_samples(self) -> enumerate[SampleType | ExistingSample]:
        return enumerate(self.samples)

    @property
    def enumerated_new_samples(self) -> list[tuple[int, SampleType]]:
        samples: list[tuple[int, SampleType]] = []
        for sample_index, sample in self.enumerated_samples:
            if not isinstance(sample, ExistingSample):
                samples.append((sample_index, sample))
        return samples

    @property
    def enumerated_existing_samples(self) -> list[tuple[int, ExistingSample]]:
        samples: list[tuple[int, ExistingSample]] = []
        for sample_index, sample in self.enumerated_samples:
            if isinstance(sample, ExistingSample):
                samples.append((sample_index, sample))
        return samples

    def get_new_sample(self, sample_name: str) -> SampleType | None:
        for _, sample in self.enumerated_new_samples:
            if sample.name == sample_name:
                return sample

    def get_existing_sample_from_db(self, sample_name: str, store: Store) -> DbSample | None:
        for _, sample in self.enumerated_existing_samples:
            db_sample: DbSample | None = store.get_sample_by_internal_id(sample.internal_id)
            if db_sample and db_sample.name == sample_name:
                return db_sample

    @model_validator(mode="before")
    def convert_empty_strings_to_none(cls, data):
        if isinstance(data, dict):
            for key, value in data.items():
                if value == "":
                    data[key] = None
        return data

    @model_validator(mode="after")
    def set_case_name_on_new_samples(self):
        """Sets the case name on new samples, so it can be easily fetched when stored in LIMS."""
        for _, sample in self.enumerated_new_samples:
            sample._case_name = self.name
        return self

    @model_validator(mode="after")
    def set_case_priority_on_new_samples(self):
        """Sets the priority on new samples, so it can be easily fetched when stored in LIMS."""
        for _, sample in self.enumerated_new_samples:
            sample._case_priority = self.priority
        return self



import logging
from datetime import datetime

from cg.constants import DataDelivery, Sex
from cg.constants.constants import Workflow
from cg.models.orders.sample_base import PriorityEnum, StatusEnum
from cg.services.orders.lims_service.service import OrderLimsService
from cg.services.orders.storing.service import StoreOrderService
from cg.services.orders.validation.order_types.taxprofiler.models.order import TaxprofilerOrder
from cg.services.orders.validation.order_types.taxprofiler.models.sample import TaxprofilerSample
from cg.store.models import ApplicationVersion
from cg.store.models import Case as DbCase
from cg.store.models import CaseSample, Customer
from cg.store.models import Order as DbOrder
from cg.store.models import Sample as DbSample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class StoreTaxprofilerOrderService(StoreOrderService):
    """Storing service for Taxprofiler orders."""

    def __init__(self, status_db: Store, lims_service: OrderLimsService):
        self.status_db = status_db
        self.lims = lims_service

    def store_order(self, order: TaxprofilerOrder) -> dict:
        """Submit a batch of taxprofiler samples."""
        project_data, lims_map = self.lims.process_lims(
            samples=order.samples,
            customer=order.customer,
            ticket=order._generated_ticket_id,
            order_name=order.name,
            workflow=Workflow.TAXPROFILER,
            delivery_type=DataDelivery(order.delivery_type),
            skip_reception_control=order.skip_reception_control,
        )
        self._fill_in_sample_ids(samples=order.samples, lims_map=lims_map)
        new_samples = self.store_order_data_in_status_db(order)
        return {"project": project_data, "records": new_samples}

    def store_order_data_in_status_db(
        self,
        order: TaxprofilerOrder,
    ) -> list[DbSample]:
        """Store samples in the StatusDB database."""
        new_samples = []
        customer: Customer = self.status_db.get_customer_by_internal_id(order.customer)
        db_order: DbOrder = self.status_db.add_order(
            customer=customer, ticket_id=order._generated_ticket_id
        )
        priority: PriorityEnum = order.samples[0].priority
        db_case: DbCase = self._create_db_case(order=order, customer=customer, priority=priority)
        db_order.cases.append(db_case)
        with self.status_db.session.no_autoflush:
            for sample in order.samples:
                db_sample: DbSample = self._create_db_sample(
                    order=order, sample=sample, customer=customer
                )
                new_relationship: CaseSample = self.status_db.relate_sample(
                    case=db_case,
                    sample=db_sample,
                    status=StatusEnum.unknown,
                    should_deliver_sample=True,
                )
                self.status_db.add_item_to_store(new_relationship)
                new_samples.append(db_sample)
        self.status_db.add_item_to_store(db_case)
        self.status_db.add_item_to_store(db_order)
        self.status_db.add_multiple_items_to_store(new_samples)
        self.status_db.commit_to_store()
        return new_samples

    def _create_db_case(
        self, order: TaxprofilerOrder, customer: Customer, priority: PriorityEnum
    ) -> DbCase:
        db_case: DbCase = self.status_db.add_case(
            data_analysis=Workflow.TAXPROFILER,
            data_delivery=DataDelivery(order.delivery_type),
            name=str(order._generated_ticket_id),
            priority=priority,
            ticket=str(order._generated_ticket_id),
        )
        db_case.customer = customer
        return db_case

    def _create_db_sample(
        self, sample: TaxprofilerSample, order: TaxprofilerOrder, customer: Customer
    ) -> DbSample:
        db_sample: DbSample = self.status_db.add_sample(
            name=sample.name,
            sex=Sex.UNKNOWN,
            comment=sample.comment,
            control=sample.control,
            internal_id=sample._generated_lims_id,
            order=order.name,
            ordered=datetime.now(),
            original_ticket=order._generated_ticket_id,
            priority=sample.priority,
        )
        application_version: ApplicationVersion = (
            self.status_db.get_current_application_version_by_tag(sample.application)
        )
        db_sample.customer = customer
        db_sample.application_version = application_version
        return db_sample



"""Chanjo2 API pydantic models."""

from pydantic import BaseModel, RootModel, field_validator


class CoverageSample(BaseModel):
    """Chanjo2 sample model defined with an ID and a coverage file path."""

    coverage_file_path: str
    name: str


class CoveragePostRequest(BaseModel):
    """Chanjo2 coverage POST request model."""

    build: str
    coverage_threshold: int
    hgnc_gene_ids: list[int]
    interval_type: str
    samples: list[CoverageSample]


class CoverageMetrics(BaseModel):
    """Chanjo2 sample coverage metrics."""

    coverage_completeness_percent: float
    mean_coverage: float


class CoverageMetricsChanjo1(CoverageMetrics):
    """Chanjo1 sample coverage metrics model."""

    pass


class CoveragePostResponse(RootModel):
    """Coverage sample data model returned from the POST request."""

    root: dict[str, CoverageMetrics]

    @field_validator("root")
    @classmethod
    def root_must_not_be_empty(cls, root: dict[str, CoverageMetrics]):
        if not root:
            raise ValueError("Coverage POST response must not be an empty dictionary")
        return root

    def get_sample_coverage_metrics(self, sample_id: str) -> CoverageMetrics:
        """Return the coverage metrics for the specified sample ID."""
        if sample_id not in self.root:
            raise ValueError(f"Sample ID '{sample_id}' not found in the coverage POST response")
        return self.root[sample_id]



from datetime import datetime

from pydantic import BaseModel


class PacbioSmrtCellMetricsDTO(BaseModel):
    barcoded_hifi_mean_read_length: int
    barcoded_hifi_reads: int
    barcoded_hifi_reads_percentage: float
    completed_at: datetime
    hifi_mean_read_length: int
    hifi_median_read_quality: str
    hifi_reads: int
    hifi_yield: int
    internal_id: str
    movie_name: str
    p0_percent: float
    p1_percent: float
    p2_percent: float
    percent_reads_passing_q30: float
    plate: int
    run_id: str
    started_at: datetime
    well: str


class PacbioSmrtCellMetricsResponse(BaseModel):
    metrics: list[PacbioSmrtCellMetricsDTO]


class PacbioSequencingRunDTO(BaseModel):
    id: int
    comment: str
    processed: bool
    run_id: str
    run_name: str | None
    unique_id: str | None


class PacbioSequencingRunResponse(BaseModel):
    pacbio_sequencing_runs: list[PacbioSequencingRunDTO]
    total_count: int


class PacbioSequencingRunUpdateRequest(BaseModel):
    id: int
    comment: str | None = None
    processed: bool | None = None



"""Module to parse the illumina demultiplexing log file."""

from pathlib import Path

from pydantic import BaseModel

from cg.io.json import read_json


class DemuxSoftware(BaseModel):
    software_version: str
    software: str


def _parse_dragen_replay(content: dict) -> DemuxSoftware:
    """Parse a dragen-replay.json file."""
    system_info: dict = content["system"]
    return DemuxSoftware(
        software_version=system_info["dragen_version"],
        software="dragen",
    )


def _parse_highlevel_summary(content: dict) -> DemuxSoftware:
    """Parse a highlevel_summary.json file."""
    return DemuxSoftware(
        software_version=content["software_version"],
        software="dragen",
    )


class IlluminaDemuxVersionService:
    @staticmethod
    def _parse_demux_version_file(demux_version_file: Path) -> DemuxSoftware:
        """Parse the demultiplexing log file."""
        content: dict = read_json(demux_version_file)
        if demux_version_file.name == "highlevel_summary.json":
            return _parse_highlevel_summary(content)
        return _parse_dragen_replay(content)

    def get_demux_software(self, demux_version_file: Path) -> str:
        """Get the demux software."""
        demux_software: DemuxSoftware = self._parse_demux_version_file(demux_version_file)
        return demux_software.software

    def get_demux_software_version(self, demux_version_file: Path) -> str:
        """Get the demultiplexing software version."""
        demux_software: DemuxSoftware = self._parse_demux_version_file(demux_version_file)
        return demux_software.software_version



import logging
from glob import glob
from pathlib import Path

import rich_click as click
from pydantic import ValidationError

from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.demultiplex.sample_sheet.api import IlluminaSampleSheetService
from cg.apps.tb import TrailblazerAPI
from cg.cli.demultiplex.copy_novaseqx_demultiplex_data import (
    create_manifest_file,
    is_flow_cell_sync_confirmed,
    is_manifest_file_required,
    is_ready_to_copy_to_demultiplexed_runs,
    is_syncing_complete,
    link_onboard_demultiplexed_flow_cell,
)
from cg.constants.cli_options import DRY_RUN
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.exc import CgError, FlowCellError, SampleSheetContentError
from cg.models.cg_config import CGConfig
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData

LOG = logging.getLogger(__name__)


@click.command(name="all")
@click.option("--sequencing-runs-directory", type=click.Path(exists=True, file_okay=False))
@DRY_RUN
@click.pass_obj
def demultiplex_all(context: CGConfig, sequencing_runs_directory: click.Path, dry_run: bool):
    """Demultiplex all sequencing runs that are ready under the sequencing runs directory."""
    LOG.info("Running cg demultiplex all ...")
    sample_sheet_api: IlluminaSampleSheetService = context.sample_sheet_api
    demultiplex_api: DemultiplexingAPI = context.demultiplex_api
    demultiplex_api.set_dry_run(dry_run=dry_run)
    if sequencing_runs_directory:
        sequencing_runs_directory: Path = Path(str(sequencing_runs_directory))
    else:
        sequencing_runs_directory: Path = Path(demultiplex_api.sequencing_runs_dir)

    LOG.info(f"Search for sequencing run ready to demultiplex in {sequencing_runs_directory}")
    for sub_dir in sequencing_runs_directory.iterdir():
        if not sub_dir.is_dir():
            continue
        LOG.info(f"Found directory {sub_dir}")
        try:
            sequencing_run = IlluminaRunDirectoryData(sequencing_run_path=sub_dir)
        except FlowCellError:
            continue

        if not demultiplex_api.is_demultiplexing_possible(sequencing_run=sequencing_run):
            LOG.warning(f"Can not start demultiplexing for sequencing run {sequencing_run.id}!")
            continue

        try:
            sample_sheet_api.validate_sample_sheet(sequencing_run.sample_sheet_path)
        except (CgError, ValidationError):
            LOG.warning(
                "Malformed sample sheet. Run "
                f"cg demultiplex samplesheet validate {sequencing_run.sample_sheet_path}"
            )
            continue

        if not dry_run:
            demultiplex_api.prepare_output_directory(sequencing_run)
            slurm_job_id: int = demultiplex_api.start_demultiplexing(sequencing_run=sequencing_run)
            tb_api: TrailblazerAPI = context.trailblazer_api
            demultiplex_api.add_to_trailblazer(
                tb_api=tb_api, slurm_job_id=slurm_job_id, sequencing_run=sequencing_run
            )


@click.command(name="sequencing-run")
@click.argument("sequencing-run-name")
@DRY_RUN
@click.pass_obj
def demultiplex_sequencing_run(
    context: CGConfig,
    dry_run: bool,
    sequencing_run_name: str,
):
    """Demultiplex an Illumina sequencing run using BCLConvert.


    sequencing run name is the sequencing run directory name, e.g. '230912_A00187_1009_AHK33MDRX3'
    """

    LOG.info(f"Starting demultiplexing of sequencing run {sequencing_run_name}")
    sample_sheet_api: IlluminaSampleSheetService = context.sample_sheet_api
    demultiplex_api: DemultiplexingAPI = context.demultiplex_api
    sequencing_run_dir: Path = Path(
        context.demultiplex_api.sequencing_runs_dir, sequencing_run_name
    )
    demultiplex_api.set_dry_run(dry_run=dry_run)
    LOG.info(f"setting flow cell id to {sequencing_run_name}")
    LOG.info(f"setting demultiplexed runs dir to {demultiplex_api.demultiplexed_runs_dir}")

    try:
        sequencing_run = IlluminaRunDirectoryData(sequencing_run_dir)
    except FlowCellError as error:
        raise click.Abort from error

    if not demultiplex_api.is_demultiplexing_possible(sequencing_run=sequencing_run):
        LOG.warning("Can not start demultiplexing!")
        return

    try:
        sample_sheet_api.validate_sample_sheet(sequencing_run.sample_sheet_path)
    except SampleSheetContentError:
        LOG.warning("Starting demultiplexing a with a manually modified sample sheet")
    except (CgError, ValidationError) as error:
        LOG.warning(
            "Malformed sample sheet. "
            f"Run cg demultiplex samplesheet validate {sequencing_run.sample_sheet_path}"
        )
        raise click.Abort from error

    if dry_run:
        LOG.info(f"Would have started demultiplexing {sequencing_run_name}")
        return
    demultiplex_api.prepare_output_directory(sequencing_run)
    slurm_job_id: int = demultiplex_api.start_demultiplexing(sequencing_run=sequencing_run)
    tb_api: TrailblazerAPI = context.trailblazer_api
    demultiplex_api.add_to_trailblazer(
        tb_api=tb_api, slurm_job_id=slurm_job_id, sequencing_run=sequencing_run
    )


@click.command(name="link-onboard-demultiplexing")
@click.pass_obj
def link_onboard_demultiplexed_flow_cells(context: CGConfig):
    """Hard-link on-instrument demultiplexed NovaSeqX flow cells into demultiplexed-runs."""
    sequencing_runs_dir: Path = Path(context.run_instruments.illumina.sequencing_runs_dir)
    demultiplexed_runs_dir: Path = Path(context.run_instruments.illumina.demultiplexed_runs_dir)

    for sequencing_run_dir in sequencing_runs_dir.iterdir():
        if is_ready_to_copy_to_demultiplexed_runs(
            flow_cell_dir=sequencing_run_dir, demultiplexed_runs_dir=demultiplexed_runs_dir
        ):
            LOG.info(f"Linking {sequencing_run_dir.name} into {demultiplexed_runs_dir}")
            link_onboard_demultiplexed_flow_cell(
                sequencing_run_dir=sequencing_run_dir,
                demultiplexed_runs_dir=demultiplexed_runs_dir,
                flow_cell_id=IlluminaRunDirectoryData(sequencing_run_dir).id,
                hk_api=context.housekeeper_api,
                tb_api=context.trailblazer_api,
            )
        else:
            LOG.info(f"Flow cell {sequencing_run_dir.name} is not ready for linking, skipping.")


@click.command(name="confirm-sequencing-run-sync")
@click.option(
    "--source-directory",
    required=True,
    help="The path from where the syncing is done.",
)
@click.pass_obj
def confirm_sequencing_run_sync(context: CGConfig, source_directory: str):
    """Checks if all relevant files for the demultiplexing have been synced.
    If so it creates a CopyComplete.txt file to show that that is the case."""
    target_sequencing_runs_directory = Path(context.run_instruments.illumina.sequencing_runs_dir)
    for source_sequencing_run in Path(source_directory).iterdir():
        target_sequencing_run = Path(target_sequencing_runs_directory, source_sequencing_run.name)
        if is_flow_cell_sync_confirmed(target_sequencing_run):
            LOG.debug(f"Flow cell {source_sequencing_run} has already been confirmed, skipping.")
            continue
        if is_syncing_complete(
            source_directory=source_sequencing_run,
            target_directory=Path(target_sequencing_runs_directory, source_sequencing_run.name),
        ):
            Path(
                target_sequencing_runs_directory,
                source_sequencing_run.name,
                DemultiplexingDirsAndFiles.COPY_COMPLETE,
            ).touch()


@click.command(name="create-manifest-files")
@click.option(
    "--source-directory",
    required=True,
    help="The path from where the syncing is done.",
)
def create_manifest_files(source_directory: str):
    """Creates a file manifest for each sequencing run in the source directory."""
    for source_sequencing_run in glob(f"{source_directory}/*"):
        if is_manifest_file_required(Path(source_sequencing_run)):
            create_manifest_file(Path(source_sequencing_run))



from pathlib import Path

from cg.services.run_devices.abstract_models import RunData


class PacBioRunData(RunData):
    """Holds information on a single SMRTcell of a PacBio run."""

    full_path: Path
    run_id: str
    well_name: str
    plate: int



from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.nallo.constants import NalloDeliveryType
from cg.services.orders.validation.order_types.nallo.models.case import NalloCase
from cg.services.orders.validation.order_types.nallo.models.sample import NalloSample


class NalloOrder(OrderWithCases[NalloCase, NalloSample]):
    delivery_type: NalloDeliveryType



import logging
from pathlib import Path

from cg.apps.tb import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis
from cg.constants import Priority, Workflow
from cg.constants.tb import AnalysisType
from cg.meta.workflow.utils.utils import MAP_TO_TRAILBLAZER_PRIORITY
from cg.services.analysis_service.analysis_service import AnalysisService
from cg.services.deliver_files.deliver_files_service.error_handling import (
    handle_no_delivery_files_error,
)
from cg.services.deliver_files.file_fetcher.abstract import FetchDeliveryFilesService
from cg.services.deliver_files.file_fetcher.models import DeliveryFiles
from cg.services.deliver_files.file_formatter.destination.abstract import (
    DeliveryDestinationFormatter,
)
from cg.services.deliver_files.file_formatter.destination.models import FormattedFiles
from cg.services.deliver_files.file_mover.abstract import DestinationFilesMover
from cg.services.deliver_files.rsync.service import DeliveryRsyncService
from cg.store.exc import EntryNotFoundError
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class DeliverFilesService:
    """
    Deliver files for a case, cases in a ticket or a sample to a specified destination or upload location.
    Requires:
        - FetchDeliveryFilesService: Service to fetch the files to deliver from housekeeper
        - DestinationFilesMover: Service to move the files to the destination of delivery or upload
        - DeliveryDestinationFormatter: Service to format the files to the destination format
        - DeliveryRsyncService: Service to run rsync for the delivery
        - TrailblazerAPI: Service to interact with Trailblazer
        - AnalysisService: Service to interact with the analysis
        - Store: Store to interact with the database
    """

    def __init__(
        self,
        delivery_file_manager_service: FetchDeliveryFilesService,
        move_file_service: DestinationFilesMover,
        file_formatter_service: DeliveryDestinationFormatter,
        rsync_service: DeliveryRsyncService,
        tb_service: TrailblazerAPI,
        analysis_service: AnalysisService,
        status_db: Store,
    ):
        self.file_manager = delivery_file_manager_service
        self.file_mover = move_file_service
        self.file_formatter = file_formatter_service
        self.status_db = status_db
        self.rsync_service = rsync_service
        self.tb_service = tb_service
        self.analysis_service = analysis_service

    @handle_no_delivery_files_error
    def deliver_files_for_case(
        self, case: Case, delivery_base_path: Path, dry_run: bool = False
    ) -> None:
        """Deliver the files for a case to the customer folder.
        args:
            case: The case to deliver files for
            delivery_base_path: The base path to deliver the files to
            dry_run: Whether to perform a dry run or not
        """
        delivery_files: DeliveryFiles = self.file_manager.get_files_to_deliver(
            case_id=case.internal_id
        )
        moved_files: DeliveryFiles = self.file_mover.move_files(
            delivery_files=delivery_files, delivery_base_path=delivery_base_path
        )
        formatted_files: FormattedFiles = self.file_formatter.format_files(
            delivery_files=moved_files
        )

        folders_to_deliver: set[Path] = set(
            [formatted_file.formatted_path.parent for formatted_file in formatted_files.files]
        )
        job_id: int = self._start_rsync_job(
            case=case, dry_run=dry_run, folders_to_deliver=folders_to_deliver
        )
        self._add_trailblazer_tracking(case=case, job_id=job_id, dry_run=dry_run)

    def deliver_files_for_ticket(
        self, ticket_id: str, delivery_base_path: Path, dry_run: bool = False
    ) -> None:
        """Deliver the files for all cases in a ticket to the customer folder.
        args:
            ticket_id: The ticket id to deliver files for
            delivery_base_path: The base path to deliver the files to
            dry_run: Whether to perform a dry run or not
        """
        cases: list[Case] = self.status_db.get_cases_by_ticket_id(ticket_id)
        if not cases:
            raise EntryNotFoundError(f"No cases found for ticket {ticket_id}")
        for case in cases:
            self.deliver_files_for_case(
                case=case, delivery_base_path=delivery_base_path, dry_run=dry_run
            )

    def deliver_files_for_sample(
        self, case: Case, sample_id: str, delivery_base_path: Path, dry_run: bool = False
    ):
        """Deliver the files for a sample to the customer folder."""
        delivery_files: DeliveryFiles = self.file_manager.get_files_to_deliver(
            case_id=case.internal_id, sample_id=sample_id
        )
        moved_files: DeliveryFiles = self.file_mover.move_files(
            delivery_files=delivery_files, delivery_base_path=delivery_base_path
        )
        formatted_files: FormattedFiles = self.file_formatter.format_files(
            delivery_files=moved_files
        )
        folders_to_deliver: set[Path] = set(
            [formatted_file.formatted_path.parent for formatted_file in formatted_files.files]
        )
        job_id: int = self._start_rsync_job(
            case=case, dry_run=dry_run, folders_to_deliver=folders_to_deliver
        )
        self._add_trailblazer_tracking(case=case, job_id=job_id, dry_run=dry_run)

    def deliver_files_for_sample_no_rsync(
        self, case: Case, sample_id: str, delivery_base_path: Path
    ):
        """
        Deliver the files for a sample to the delivery base path. Does not perform rsync.
        args:
            case: The case to deliver files for
            sample_id: The sample to deliver files for
            delivery_base_path: The base path to deliver the files to
        """
        delivery_files: DeliveryFiles = self.file_manager.get_files_to_deliver(
            case_id=case.internal_id, sample_id=sample_id
        )
        moved_files: DeliveryFiles = self.file_mover.move_files(
            delivery_files=delivery_files, delivery_base_path=delivery_base_path
        )
        self.file_formatter.format_files(delivery_files=moved_files)

    def _start_rsync_job(self, case: Case, dry_run: bool, folders_to_deliver: set[Path]) -> int:
        """Start a rsync job for the case.
        args:
            case: The case to start the rsync job for
            dry_run: Whether to perform a dry run or not
            folders_to_deliver: The folders to deliver
        """
        LOG.debug(f"[RSYNC] Starting rsync job for case {case.internal_id}")
        job_id: int = self.rsync_service.run_rsync_for_case(
            case=case,
            dry_run=dry_run,
            folders_to_deliver=folders_to_deliver,
        )
        self.rsync_service.write_trailblazer_config(
            content={"jobs": [str(job_id)]},
            config_path=self.rsync_service.trailblazer_config_path,
            dry_run=dry_run,
        )
        return job_id

    def _add_trailblazer_tracking(self, case: Case, job_id: int, dry_run: bool) -> None:
        """Add the rsync job to Trailblazer for tracking.
        args:
            case: The case to add the job for
            job_id: The job id to add for trailblazer tracking
            dry_run: Whether to perform a dry run or not
        """
        if dry_run:
            LOG.info(f"Would have added the analysis for case {case.internal_id} to Trailblazer")
        else:
            LOG.debug(f"[TB SERVICE] Adding analysis for case {case.internal_id} to Trailblazer")
            analysis: TrailblazerAnalysis = self.tb_service.add_pending_analysis(
                case_id=f"{case.internal_id}_rsync",
                analysis_type=AnalysisType.OTHER,
                config_path=self.rsync_service.trailblazer_config_path.as_posix(),
                order_id=case.latest_order.id,
                out_dir=self.rsync_service.log_dir.as_posix(),
                priority=MAP_TO_TRAILBLAZER_PRIORITY[case.priority],
                workflow=Workflow.RSYNC,
                ticket=case.latest_ticket,
            )
            self.tb_service.add_upload_job_to_analysis(analysis_id=analysis.id, slurm_id=job_id)
            self.analysis_service.add_upload_job(slurm_id=job_id, case_id=case.internal_id)
            LOG.info(f"Transfer of case {case.internal_id} started with SLURM job id {job_id}")



from cg.services.orders.validation.constants import MAXIMUM_VOLUME, MINIMUM_VOLUME
from cg.services.orders.validation.errors.case_errors import CaseError
from cg.services.orders.validation.errors.sample_errors import SampleError


class CaseSampleError(CaseError, SampleError):
    pass


class OccupiedWellError(CaseSampleError):
    field: str = "well_position"
    message: str = "Well is already occupied"


class ApplicationArchivedError(CaseSampleError):
    field: str = "application"
    message: str = "Chosen application is archived"


class ApplicationNotValidError(CaseSampleError):
    field: str = "application"
    message: str = "Chosen application does not exist"


class ApplicationNotCompatibleError(CaseSampleError):
    field: str = "application"
    message: str = "Application is not allowed for the chosen workflow"


class SampleNameRepeatedError(CaseSampleError):
    field: str = "name"
    message: str = "Sample name already used"


class SampleNameSameAsCaseNameError(CaseSampleError):
    field: str = "name"
    message: str = "Sample name can not be the same as any case name in order"


class InvalidFatherSexError(CaseSampleError):
    field: str = "father"
    message: str = "Father must be male"


class FatherNotInCaseError(CaseSampleError):
    field: str = "father"
    message: str = "Father must be in the same case"


class InvalidMotherSexError(CaseSampleError):
    field: str = "mother"
    message: str = "Mother must be female"


class PedigreeError(CaseSampleError):
    message: str = "Invalid pedigree relationship"


class DescendantAsMotherError(PedigreeError):
    field: str = "mother"
    message: str = "Descendant sample cannot be mother"


class DescendantAsFatherError(PedigreeError):
    field: str = "father"
    message: str = "Descendant sample cannot be father"


class SampleIsOwnMotherError(PedigreeError):
    field: str = "mother"
    message: str = "Sample cannot be its own mother"


class SampleIsOwnFatherError(PedigreeError):
    field: str = "father"
    message: str = "Sample cannot be its own father"


class MotherNotInCaseError(CaseSampleError):
    field: str = "mother"
    message: str = "Mother must be in the same case"


class SampleDoesNotExistError(CaseSampleError):
    field: str = "internal_id"
    message: str = "The sample does not exist"


class SubjectIdSameAsCaseNameError(CaseSampleError):
    field: str = "subject_id"
    message: str = "Subject id must be different from the case name"


class ConcentrationRequiredIfSkipRCError(CaseSampleError):
    field: str = "concentration_ng_ul"
    message: str = "Concentration is required when skipping reception control"


class SubjectIdSameAsSampleNameError(CaseSampleError):
    field: str = "subject_id"
    message: str = "Subject id must be different from the sample name"


class InvalidConcentrationIfSkipRCError(CaseSampleError):
    def __init__(self, case_index: int, sample_index: int, allowed_interval: tuple[float, float]):
        field: str = "concentration_ng_ul"
        message: str = (
            f"Concentration must be between {allowed_interval[0]} ng/μL and {allowed_interval[1]} ng/μL if reception control should be skipped"
        )
        super(CaseSampleError, self).__init__(
            case_index=case_index, sample_index=sample_index, field=field, message=message
        )


class WellPositionMissingError(CaseSampleError):
    field: str = "well_position"
    message: str = "Well position is required for well plates"


class ContainerNameMissingError(CaseSampleError):
    field: str = "container_name"
    message: str = "Container name is required for well plates"


class InvalidVolumeError(CaseSampleError):
    field: str = "volume"
    message: str = f"Volume must be between {MINIMUM_VOLUME}-{MAXIMUM_VOLUME} μL"


class VolumeRequiredError(CaseSampleError):
    field: str = "volume"
    message: str = "Volume is required"


class InvalidBufferError(CaseSampleError):
    field: str = "elution_buffer"
    message: str = "The chosen buffer is not allowed when skipping reception control"


class SexSubjectIdError(CaseSampleError):
    field: str = "sex"
    message: str = "Another sample with the same subject id has a different sex"


class CaptureKitResetError(CaseSampleError):
    field: str = "warnings"
    message: str = "No bait set will be used, since it is not required for this application."


class CaptureKitMissingError(CaseSampleError):
    field: str = "capture_kit"
    message: str = "Bait set is required for TGS analyses"


class InvalidCaptureKitError(CaseSampleError):
    field: str = "capture_kit"
    message: str = "Bait set must be valid"


class WellFormatError(CaseSampleError):
    field: str = "well_position"
    message: str = "Well position must follow the format A-H:1-12"


class ContainerNameRepeatedError(CaseSampleError):
    field: str = "container_name"
    message: str = "Tube names must be unique among samples"


class StatusUnknownError(CaseSampleError):
    field: str = "status"
    message: str = "Samples in case cannot all have status unknown"


class BufferMissingError(CaseSampleError):
    field: str = "elution_buffer"
    message: str = "Buffer must be specified with this application"


class SampleOutsideOfCollaborationError(CaseSampleError):
    field: str = "internal_id"
    message: str = "Sample cannot be outside of collaboration"


class ExistingSampleWrongTypeError(CaseSampleError):
    field: str = "internal_id"
    message: str = "The sample is not compatible with the order type."


class SampleNameAlreadyExistsError(CaseSampleError):
    field: str = "name"
    message: str = "Sample name already exists in a previous order"


class MissingSourceCommentError(CaseSampleError):
    field: str = "source_comment"
    message: str = "Other source must be filled in when source is 'other'"


class NormalSampleNotAllowedError(CaseSampleError):
    field: str = "internal_id"
    message: str = "Existing samples must be tumour samples"


class TumourValueResetError(CaseSampleError):
    field: str = "warnings"
    message: str = "The tumour status was overridden to True. Required for RNAFusion analysis."



import logging
from pathlib import Path

LOG = logging.getLogger(__name__)


def validate_name_pre_fix(run_name: str, pre_fix: str) -> None:
    if not run_name.startswith(pre_fix):
        raise ValueError(f"Run name does not start with {pre_fix}")


def validate_has_expected_parts(run_name: str, expected_parts: int) -> None:
    if len(run_name.split("/")) != expected_parts:
        raise ValueError(f"Run name does not have number of expected parts {expected_parts}.")


def validate_files_or_directories_exist(files: list[Path]) -> None:
    """Validate that all files or directories in the list exist."""
    exit_success: bool = True
    for file in files:
        if not file.exists():
            LOG.error(f"File or directory {file.as_posix()} does not exist")
            exit_success = False
    if not exit_success:
        raise FileNotFoundError("Some of the provided paths do not exist")



from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.balsamic_umi.constants import BalsamicUmiDeliveryType
from cg.services.orders.validation.order_types.balsamic_umi.models.case import BalsamicUmiCase
from cg.services.orders.validation.order_types.balsamic_umi.models.sample import BalsamicUmiSample


class BalsamicUmiOrder(OrderWithCases[BalsamicUmiCase, BalsamicUmiSample]):
    delivery_type: BalsamicUmiDeliveryType



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.fluffy.constants import FluffyDeliveryType
from cg.services.orders.validation.order_types.fluffy.models.sample import FluffySample


class FluffyOrder(OrderWithSamples[FluffySample]):
    delivery_type: FluffyDeliveryType

    @property
    def pools(self) -> dict[str, list[FluffySample]]:
        """Return a dictionary matching pool names and their respective samples."""
        pools: dict[str, list[FluffySample]] = {}
        for sample in self.samples:
            if sample.pool not in pools:
                pools[sample.pool] = [sample]
            else:
                pools[sample.pool].append(sample)
        return pools

    @property
    def enumerated_pools(self) -> dict[str, list[tuple[int, FluffySample]]]:
        """Return the pool dictionary with indexes for the samples to map them to validation errors."""
        pools: dict[str, list[tuple[int, FluffySample]]] = {}
        for sample_index, sample in self.enumerated_samples:
            if sample.pool not in pools:
                pools[sample.pool] = [(sample_index, sample)]
            else:
                pools[sample.pool].append((sample_index, sample))
        return pools
